import pytest

from dagster.core.errors import DagsterRuntimeCoercionError

from dagster.core.types import Int, Nullable, List, PythonObjectType
from dagster.core.types.runtime import resolve_to_runtime_type


class BarObj(object):
    pass


class Bar(PythonObjectType):
    def __init__(self):
        super(Bar, self).__init__(BarObj, description='A bar.')


def test_python_object_type():
    type_bar = Bar.inst()

    assert type_bar.name == 'Bar'
    assert type_bar.description == 'A bar.'
    assert type_bar.coerce_runtime_value(BarObj())

    with pytest.raises(DagsterRuntimeCoercionError):
        assert type_bar.coerce_runtime_value(None)
    with pytest.raises(DagsterRuntimeCoercionError):
        type_bar.coerce_runtime_value('not_a_bar')


def test_nullable_python_object_type():
    nullable_type_bar = resolve_to_runtime_type(Nullable(Bar))

    assert nullable_type_bar.coerce_runtime_value(BarObj())
    assert nullable_type_bar.coerce_runtime_value(None) is None

    with pytest.raises(DagsterRuntimeCoercionError):
        nullable_type_bar.coerce_runtime_value('not_a_bar')


def test_nullable_int_coercion():
    int_type = resolve_to_runtime_type(Int)
    assert int_type.coerce_runtime_value(1) == 1

    with pytest.raises(DagsterRuntimeCoercionError):
        assert int_type.coerce_runtime_value(None)

    nullable_int_type = resolve_to_runtime_type(Nullable(Int))
    assert nullable_int_type.coerce_runtime_value(1) == 1
    assert nullable_int_type.coerce_runtime_value(None) is None


def assert_success(fn, value):
    assert fn(value) == value


def assert_failure(fn, value):
    with pytest.raises(DagsterRuntimeCoercionError):
        fn(value)


def test_nullable_list_combos_coerciion():

    list_of_int = resolve_to_runtime_type(List(Int))

    assert_failure(list_of_int.coerce_runtime_value, None)
    assert_success(list_of_int.coerce_runtime_value, [])
    assert_success(list_of_int.coerce_runtime_value, [1])
    assert_failure(list_of_int.coerce_runtime_value, [None])

    nullable_int_of_list = resolve_to_runtime_type(Nullable(List(Int)))

    assert_success(nullable_int_of_list.coerce_runtime_value, None)
    assert_success(nullable_int_of_list.coerce_runtime_value, [])
    assert_success(nullable_int_of_list.coerce_runtime_value, [1])
    assert_failure(nullable_int_of_list.coerce_runtime_value, [None])

    list_of_nullable_int = resolve_to_runtime_type(List(Nullable(Int)))
    assert_failure(list_of_nullable_int.coerce_runtime_value, None)
    assert_success(list_of_nullable_int.coerce_runtime_value, [])
    assert_success(list_of_nullable_int.coerce_runtime_value, [1])
    assert_success(list_of_nullable_int.coerce_runtime_value, [None])

    nullable_list_of_nullable_int = resolve_to_runtime_type(Nullable(List(Nullable(Int))))
    assert_success(nullable_list_of_nullable_int.coerce_runtime_value, None)
    assert_success(nullable_list_of_nullable_int.coerce_runtime_value, [])
    assert_success(nullable_list_of_nullable_int.coerce_runtime_value, [1])
    assert_success(nullable_list_of_nullable_int.coerce_runtime_value, [None])



from dagster import check

from dagster.core.types.field import Field

from .input import InputDefinition

from .output import OutputDefinition

from .utils import check_valid_name


class SolidDefinition(object):
    '''A solid (a name extracted from the acronym of "software-structured data" (SSD)) represents
    a unit of computation within a data pipeline.

    As its core, a solid is a function. It accepts inputs (which are values produced from
    other solids) and configuration, and produces outputs. These solids are composed as a
    directed, acyclic graph (DAG) within a pipeline to form a computation that produces
    data assets.

    Solids should be implemented as idempotent, parameterizable, non-destructive functions.
    Data computations with these properties are much easier to test, reason about, and operate.

    The inputs and outputs are gradually, optionally typed by the dagster type system. Types
    can be user-defined and can represent entites as varied as scalars, dataframe, database
    tables, and so forth. They can represent pure in-memory objects, or handles to assets
    on disk or in external resources.

    A solid is a generalized abstraction that could take many forms.

    Example:

        .. code-block:: python

            def _read_csv(info, inputs):
                yield Result(pandas.read_csv(info.config['path']))

            SolidDefinition(
                name='read_csv',
                inputs=[],
                config_field=Field(types.Dict({'path' => types.Path})),
                outputs=[OutputDefinition()] # default name ('result') and any typed
                transform_fn
            )

    Attributes:
        name (str): Name of the solid.
        input_defs (List[InputDefinition]): Inputs of the solid.
        transform_fn (callable):
            Callable with the signature
            (
                info: TransformExecutionInfo,
                inputs: Dict[str, Any],
            ) : Iterable<Result>
        outputs_defs (List[OutputDefinition]): Outputs of the solid.
        config_field (Field): How the solid configured.
        description (str): Description of the solid.
        metadata (dict):
            Arbitrary metadata for the solid. Some frameworks expect and require
            certain metadata to be attached to a solid.
    '''

    def __init__(
        self,
        name,
        inputs,
        transform_fn,
        outputs,
        config_field=None,
        description=None,
        metadata=None,
    ):
        self.name = check_valid_name(name)
        self.input_defs = check.list_param(inputs, 'inputs', InputDefinition)
        self.transform_fn = check.callable_param(transform_fn, 'transform_fn')
        self.output_defs = check.list_param(outputs, 'outputs', OutputDefinition)
        self.description = check.opt_str_param(description, 'description')
        self.config_field = check.opt_inst_param(config_field, 'config_field', Field)
        self.metadata = check.opt_dict_param(metadata, 'metadata', key_type=str)
        self._input_dict = {inp.name: inp for inp in inputs}
        self._output_dict = {output.name: output for output in outputs}

    def has_input(self, name):
        check.str_param(name, 'name')
        return name in self._input_dict

    def input_def_named(self, name):
        check.str_param(name, 'name')
        return self._input_dict[name]

    def has_output(self, name):
        check.str_param(name, 'name')
        return name in self._output_dict

    def output_def_named(self, name):
        check.str_param(name, 'name')
        return self._output_dict[name]



from dagster.utils.logging import define_structured_logger, DEBUG

from dagster.core.events import (
    EventType,
    construct_event_record,
    LogMessageRecord,
    PipelineEventRecord,
)

from dagster.utils.test import create_test_runtime_execution_context


def test_structured_logger_in_context():
    messages = []

    def _append_message(logger_message):
        messages.append(logger_message)

    logger = define_structured_logger('some_name', _append_message, level=DEBUG)
    context = create_test_runtime_execution_context(loggers=[logger])
    context.debug('from_context', foo=2)
    assert len(messages) == 1
    message = messages[0]
    assert message.name == 'some_name'
    assert message.level == DEBUG
    assert message.meta['foo'] == 2
    assert message.meta['orig_message'] == 'from_context'


def test_construct_event_record():
    messages = []

    def _append_message(logger_message):
        messages.append(construct_event_record(logger_message))

    logger = define_structured_logger('some_name', _append_message, level=DEBUG)
    context = create_test_runtime_execution_context(loggers=[logger])
    context.info('random message')

    assert len(messages) == 1
    message = messages[0]
    assert isinstance(message, LogMessageRecord)

    with context.value('pipeline', 'some_pipeline'):
        context.events.pipeline_start()

    assert len(messages) == 2
    pipeline_start = messages[1]
    assert isinstance(pipeline_start, PipelineEventRecord)
    assert pipeline_start.event_type == EventType.PIPELINE_START



from dagster import lambda_solid, PipelineDefinition, RepositoryDefinition


@lambda_solid
def hello_world():
    pass


def define_repo_demo_pipeline():
    return PipelineDefinition(name='repo_demo_pipeline', solids=[hello_world])


def define_repo():
    return RepositoryDefinition(
        name='demo_repository',
        pipeline_dict={'repo_demo_pipeline': define_repo_demo_pipeline},
    )



from collections import namedtuple

import pytest

from dagster import (
    Any,
    DagsterEvaluateConfigValueError,
    Dict,
    ExecutionContext,
    Field,
    Int,
    List,
    Nullable,
    PipelineConfigEvaluationError,
    PipelineContextDefinition,
    PipelineDefinition,
    SolidDefinition,
    String,
    execute_pipeline,
    solid,
)


from dagster.core.types.evaluator import evaluate_config_value, DagsterEvaluationErrorReason

from dagster.core.test_utils import throwing_evaluate_config_value


def test_noop_config():
    assert Field(Any)


def test_int_field():
    config_field = Field(Dict({'int_field': Field(Int)}))

    assert evaluate_config_value(config_field.config_type, {'int_field': 1}).value == {
        'int_field': 1
    }


def assert_config_value_success(config_type, config_value, expected):
    result = evaluate_config_value(config_type, config_value)
    assert result.success
    assert result.value == expected


def assert_eval_failure(config_type, value):
    assert not evaluate_config_value(config_type, value).success


def test_int_fails():
    config_field = Field(Dict({'int_field': Field(Int)}))

    assert_eval_failure(config_field.config_type, {'int_field': 'fjkdj'})
    assert_eval_failure(config_field.config_type, {'int_field': True})


def test_default_arg():
    config_field = Field(Dict({'int_field': Field(Int, default_value=2, is_optional=True)}))

    assert_config_value_success(config_field.config_type, {}, {'int_field': 2})


def _single_required_string_config_dict():
    return Field(Dict({'string_field': Field(String)}))


def _multiple_required_fields_config_dict():
    return Field(Dict({'field_one': Field(String), 'field_two': Field(String)}))


def _single_optional_string_config_dict():
    return Field(Dict({'optional_field': Field(String, is_optional=True)}))


def _single_optional_string_field_config_dict_with_default():
    optional_field_def = Field(String, is_optional=True, default_value='some_default')
    return Field(Dict({'optional_field': optional_field_def}))


def _mixed_required_optional_string_config_dict_with_default():
    return Field(
        Dict(
            {
                'optional_arg': Field(String, is_optional=True, default_value='some_default'),
                'required_arg': Field(String, is_optional=False),
                'optional_arg_no_default': Field(String, is_optional=True),
            }
        )
    )


def _validate(config_field, value):
    return throwing_evaluate_config_value(config_field.config_type, value)


def test_single_required_string_field_config_type():
    assert _validate(_single_required_string_config_dict(), {'string_field': 'value'}) == {
        'string_field': 'value'
    }

    with pytest.raises(DagsterEvaluateConfigValueError):
        _validate(_single_required_string_config_dict(), {})

    with pytest.raises(DagsterEvaluateConfigValueError):
        _validate(_single_required_string_config_dict(), {'extra': 'yup'})

    with pytest.raises(DagsterEvaluateConfigValueError):
        _validate(_single_required_string_config_dict(), {'string_field': 'yupup', 'extra': 'yup'})

    with pytest.raises(DagsterEvaluateConfigValueError):
        _validate(_single_required_string_config_dict(), {'string_field': 1})


def test_multiple_required_fields_passing():
    assert _validate(
        _multiple_required_fields_config_dict(),
        {'field_one': 'value_one', 'field_two': 'value_two'},
    ) == {'field_one': 'value_one', 'field_two': 'value_two'}


def test_multiple_required_fields_failing():
    with pytest.raises(DagsterEvaluateConfigValueError):
        _validate(_multiple_required_fields_config_dict(), {})

    with pytest.raises(DagsterEvaluateConfigValueError):
        _validate(_multiple_required_fields_config_dict(), {'field_one': 'yup'})

    with pytest.raises(DagsterEvaluateConfigValueError):
        _validate(_multiple_required_fields_config_dict(), {'field_one': 'yup', 'extra': 'yup'})

    with pytest.raises(DagsterEvaluateConfigValueError):
        _validate(
            _multiple_required_fields_config_dict(), {'field_one': 'value_one', 'field_two': 2}
        )


def test_single_optional_field_passing():
    assert _validate(_single_optional_string_config_dict(), {'optional_field': 'value'}) == {
        'optional_field': 'value'
    }
    assert _validate(_single_optional_string_config_dict(), {}) == {}

    with pytest.raises(DagsterEvaluateConfigValueError):
        assert _validate(_single_optional_string_config_dict(), {'optional_field': None}) == {
            'optional_field': None
        }


def test_single_optional_field_failing():
    with pytest.raises(DagsterEvaluateConfigValueError):

        _validate(_single_optional_string_config_dict(), {'optional_field': 1})

    with pytest.raises(DagsterEvaluateConfigValueError):
        _validate(_single_optional_string_config_dict(), {'dlkjfalksdjflksaj': 1})


def test_single_optional_field_passing_with_default():
    assert _validate(_single_optional_string_field_config_dict_with_default(), {}) == {
        'optional_field': 'some_default'
    }

    assert _validate(
        _single_optional_string_field_config_dict_with_default(), {'optional_field': 'override'}
    ) == {'optional_field': 'override'}


def test_mixed_args_passing():
    assert _validate(
        _mixed_required_optional_string_config_dict_with_default(),
        {'optional_arg': 'value_one', 'required_arg': 'value_two'},
    ) == {'optional_arg': 'value_one', 'required_arg': 'value_two'}

    assert _validate(
        _mixed_required_optional_string_config_dict_with_default(), {'required_arg': 'value_two'}
    ) == {'optional_arg': 'some_default', 'required_arg': 'value_two'}

    assert _validate(
        _mixed_required_optional_string_config_dict_with_default(),
        {'required_arg': 'value_two', 'optional_arg_no_default': 'value_three'},
    ) == {
        'optional_arg': 'some_default',
        'required_arg': 'value_two',
        'optional_arg_no_default': 'value_three',
    }


def _single_nested_config():
    return Field(Dict({'nested': Field(Dict({'int_field': Field(Int)}))}))


def _nested_optional_config_with_default():
    return Field(
        Dict({'nested': Field(Dict({'int_field': Field(Int, is_optional=True, default_value=3)}))})
    )


def _nested_optional_config_with_no_default():
    nested_type = Dict({'int_field': Field(Int, is_optional=True)})
    return Field(Dict({'nested': Field(dagster_type=nested_type)}))


def test_single_nested_config():
    assert _validate(_single_nested_config(), {'nested': {'int_field': 2}}) == {
        'nested': {'int_field': 2}
    }


def test_single_nested_config_failures():
    with pytest.raises(DagsterEvaluateConfigValueError):
        _validate(_single_nested_config(), {'nested': 'dkjfdk'})

    with pytest.raises(DagsterEvaluateConfigValueError):
        _validate(_single_nested_config(), {'nested': {'int_field': 'dkjfdk'}})

    with pytest.raises(DagsterEvaluateConfigValueError):
        _validate(_single_nested_config(), {'nested': {'int_field': {'too_nested': 'dkjfdk'}}})


def test_nested_optional_with_default():
    assert _validate(_nested_optional_config_with_default(), {'nested': {'int_field': 2}}) == {
        'nested': {'int_field': 2}
    }

    assert _validate(_nested_optional_config_with_default(), {'nested': {}}) == {
        'nested': {'int_field': 3}
    }


def test_nested_optional_with_no_default():
    assert _validate(_nested_optional_config_with_no_default(), {'nested': {'int_field': 2}}) == {
        'nested': {'int_field': 2}
    }

    assert _validate(_nested_optional_config_with_no_default(), {'nested': {}}) == {'nested': {}}


def single_elem(ddict):
    return list(ddict.items())[0]


def test_build_optionality():
    optional_test_type = Field(
        Dict(
            {
                'required': Field(Dict({'value': Field(String)})),
                'optional': Field(Dict({'value': Field(String, is_optional=True)})),
            }
        )
    ).config_type

    assert optional_test_type.fields['required'].is_optional is False
    assert optional_test_type.fields['optional'].is_optional is True


def test_wrong_solid_name():
    pipeline_def = PipelineDefinition(
        name='pipeline_wrong_solid_name',
        solids=[
            SolidDefinition(
                name='some_solid',
                inputs=[],
                outputs=[],
                config_field=Field(Int),
                transform_fn=lambda *_args: None,
            )
        ],
    )

    env_config = {'solids': {'another_name': {'config': {}}}}

    with pytest.raises(PipelineConfigEvaluationError) as pe_info:
        execute_pipeline(pipeline_def, env_config)

    pe = pe_info.value

    assert 'Undefined field "another_name" at path root:solids' in str(pe)


def fail_me():
    assert False
    return None


def test_multiple_context():
    pipeline_def = PipelineDefinition(
        name='pipeline_test_multiple_context',
        context_definitions={
            'context_one': PipelineContextDefinition(context_fn=lambda *_args: fail_me()),
            'context_two': PipelineContextDefinition(context_fn=lambda *_args: fail_me()),
        },
        solids=[],
    )

    with pytest.raises(PipelineConfigEvaluationError):
        execute_pipeline(pipeline_def, {'context': {'context_one': {}, 'context_two': {}}})


def test_wrong_context():
    pipeline_def = PipelineDefinition(
        name='pipeline_test_multiple_context',
        context_definitions={
            'context_one': PipelineContextDefinition(context_fn=lambda *_args: fail_me()),
            'context_two': PipelineContextDefinition(context_fn=lambda *_args: fail_me()),
        },
        solids=[],
    )

    with pytest.raises(
        PipelineConfigEvaluationError, match='Undefined field "nope" at path root:context'
    ):
        execute_pipeline(pipeline_def, {'context': {'nope': {}}})


def test_solid_list_config():
    value = [1, 2]
    called = {}

    def _test_config(info, _inputs):
        assert info.config == value
        called['yup'] = True

    pipeline_def = PipelineDefinition(
        name='solid_list_config_pipeline',
        solids=[
            SolidDefinition(
                name='solid_list_config',
                inputs=[],
                outputs=[],
                config_field=Field(List(Int)),
                transform_fn=_test_config,
            )
        ],
    )

    result = execute_pipeline(
        pipeline_def, environment={'solids': {'solid_list_config': {'config': value}}}
    )

    assert result.success
    assert called['yup']


def test_two_list_types():
    assert PipelineDefinition(
        name='two_types',
        solids=[
            SolidDefinition(
                name='two_list_type',
                inputs=[],
                outputs=[],
                config_field=Field(
                    Dict({'list_one': Field(List(Int)), 'list_two': Field(List(Int))})
                ),
                transform_fn=lambda *_args: None,
            )
        ],
    )


def test_multilevel_default_handling():
    @solid(config_field=Field(Int, is_optional=True, default_value=234))
    def has_default_value(info):
        assert info.config == 234

    pipeline_def = PipelineDefinition(
        name='multilevel_default_handling', solids=[has_default_value]
    )

    assert execute_pipeline(pipeline_def).success
    assert execute_pipeline(pipeline_def, environment=None).success
    assert execute_pipeline(pipeline_def, environment={}).success
    assert execute_pipeline(pipeline_def, environment={'solids': None}).success
    assert execute_pipeline(pipeline_def, environment={'solids': {}}).success
    assert execute_pipeline(
        pipeline_def, environment={'solids': {'has_default_value': None}}
    ).success

    assert execute_pipeline(pipeline_def, environment={'solids': {'has_default_value': {}}}).success

    assert execute_pipeline(
        pipeline_def, environment={'solids': {'has_default_value': {'config': 234}}}
    ).success


def test_no_env_missing_required_error_handling():
    @solid(config_field=Field(Int))
    def required_int_solid(_info):
        pass

    pipeline_def = PipelineDefinition(
        name='no_env_missing_required_error', solids=[required_int_solid]
    )

    with pytest.raises(PipelineConfigEvaluationError) as pe_info:
        execute_pipeline(pipeline_def)

    assert isinstance(pe_info.value, PipelineConfigEvaluationError)
    pe = pe_info.value
    assert len(pe.errors) == 1
    mfe = pe.errors[0]
    assert mfe.reason == DagsterEvaluationErrorReason.MISSING_REQUIRED_FIELD
    assert len(pe.error_messages) == 1

    assert 'Missing required field "solids"' in pe.message
    assert 'at document config root' in pe.message
    assert 'Missing required field "solids"' in pe.error_messages[0]


def test_root_extra_field():
    @solid(config_field=Field(Int))
    def required_int_solid(_info):
        pass

    pipeline_def = PipelineDefinition(name='root_extra_field', solids=[required_int_solid])

    with pytest.raises(PipelineConfigEvaluationError) as pe_info:
        execute_pipeline(
            pipeline_def,
            environment={'solids': {'required_int_solid': {'config': 948594}}, 'nope': None},
        )

    pe = pe_info.value
    assert len(pe.errors) == 1
    fnd = pe.errors[0]
    assert fnd.reason == DagsterEvaluationErrorReason.FIELD_NOT_DEFINED
    assert 'Undefined field "nope"' in pe.message


def test_deeper_path():
    @solid(config_field=Field(Int))
    def required_int_solid(_info):
        pass

    pipeline_def = PipelineDefinition(name='deeper_path', solids=[required_int_solid])

    with pytest.raises(PipelineConfigEvaluationError) as pe_info:
        execute_pipeline(
            pipeline_def, environment={'solids': {'required_int_solid': {'config': 'asdf'}}}
        )

    pe = pe_info.value
    assert len(pe.errors) == 1
    rtm = pe.errors[0]
    assert rtm.reason == DagsterEvaluationErrorReason.RUNTIME_TYPE_MISMATCH


def test_working_list_path():
    called = {}

    @solid(config_field=Field(List(Int)))
    def required_list_int_solid(info):
        assert info.config == [1, 2]
        called['yup'] = True

    pipeline_def = PipelineDefinition(name='list_path', solids=[required_list_int_solid])

    result = execute_pipeline(
        pipeline_def, environment={'solids': {'required_list_int_solid': {'config': [1, 2]}}}
    )

    assert result.success
    assert called['yup']


def test_item_error_list_path():
    called = {}

    @solid(config_field=Field(List(Int)))
    def required_list_int_solid(info):
        assert info.config == [1, 2]
        called['yup'] = True

    pipeline_def = PipelineDefinition(name='list_path', solids=[required_list_int_solid])

    with pytest.raises(PipelineConfigEvaluationError) as pe_info:
        execute_pipeline(
            pipeline_def,
            environment={'solids': {'required_list_int_solid': {'config': [1, 'nope']}}},
        )

    pe = pe_info.value
    assert len(pe.errors) == 1
    rtm = pe.errors[0]
    assert rtm.reason == DagsterEvaluationErrorReason.RUNTIME_TYPE_MISMATCH

    assert 'Type failure at path "root:solids:required_list_int_solid:config[1]"' in str(pe)


def test_context_selector_working():
    called = {}

    @solid
    def check_context(info):
        assert info.context.resources == 32
        called['yup'] = True

    pipeline_def = PipelineDefinition(
        name='context_selector_working',
        solids=[check_context],
        context_definitions={
            'context_required_int': PipelineContextDefinition(
                context_fn=lambda info: ExecutionContext(resources=info.config),
                config_field=Field(Int),
            )
        },
    )

    result = execute_pipeline(
        pipeline_def, environment={'context': {'context_required_int': {'config': 32}}}
    )

    assert result.success
    assert called['yup']


def test_context_selector_extra_context():
    @solid
    def check_context(_info):
        assert False

    pipeline_def = PipelineDefinition(
        name='context_selector_extra_context',
        solids=[check_context],
        context_definitions={
            'context_required_int': PipelineContextDefinition(
                context_fn=lambda info: ExecutionContext(resources=info.config),
                config_field=Field(Int),
            )
        },
    )

    with pytest.raises(PipelineConfigEvaluationError) as pe_info:
        execute_pipeline(
            pipeline_def,
            environment={
                'context': {
                    'context_required_int': {'config': 32},
                    'extra_context': {'config': None},
                }
            },
        )

    pe = pe_info.value
    cse = pe.errors[0]
    assert cse.reason == DagsterEvaluationErrorReason.SELECTOR_FIELD_ERROR
    assert 'Specified more than one field at path "root:context"' in str(pe)


def test_context_selector_wrong_name():
    @solid
    def check_context(_info):
        assert False

    pipeline_def = PipelineDefinition(
        name='context_selector_wrong_name',
        solids=[check_context],
        context_definitions={
            'context_required_int': PipelineContextDefinition(
                context_fn=lambda info: ExecutionContext(resources=info.config),
                config_field=Field(Int),
            )
        },
    )

    with pytest.raises(PipelineConfigEvaluationError) as pe_info:
        execute_pipeline(pipeline_def, environment={'context': {'wrong_name': {'config': None}}})

    pe = pe_info.value
    cse = pe.errors[0]
    assert cse.reason == DagsterEvaluationErrorReason.FIELD_NOT_DEFINED
    assert 'Undefined field "wrong_name" at path root:context' in str(pe)


def test_context_selector_none_given():
    @solid
    def check_context(_info):
        assert False

    pipeline_def = PipelineDefinition(
        name='context_selector_none_given',
        solids=[check_context],
        context_definitions={
            'context_required_int': PipelineContextDefinition(
                context_fn=lambda info: ExecutionContext(resources=info.config),
                config_field=Field(Int),
            )
        },
    )

    with pytest.raises(PipelineConfigEvaluationError) as pe_info:
        execute_pipeline(pipeline_def, environment={'context': None})

    pe = pe_info.value
    cse = pe.errors[0]
    assert cse.reason == DagsterEvaluationErrorReason.SELECTOR_FIELD_ERROR
    assert 'You specified no fields at path "root:context"' in str(pe)


def test_multilevel_good_error_handling_solids():
    @solid(config_field=Field(Int))
    def good_error_handling(_info):
        pass

    pipeline_def = PipelineDefinition(
        name='multilevel_good_error_handling', solids=[good_error_handling]
    )

    with pytest.raises(PipelineConfigEvaluationError) as pe_info:
        execute_pipeline(pipeline_def, environment={'solids': None})

    pe = pe_info.value
    assert 'Missing required field "good_error_handling" at path root:solids' in str(pe)


def test_multilevel_good_error_handling_solid_name_solids():
    @solid(config_field=Field(Int))
    def good_error_handling(_info):
        pass

    pipeline_def = PipelineDefinition(
        name='multilevel_good_error_handling', solids=[good_error_handling]
    )

    with pytest.raises(PipelineConfigEvaluationError) as pe_info:
        execute_pipeline(pipeline_def, environment={'solids': {'good_error_handling': {}}})

    pe = pe_info.value
    assert 'Missing required field "config" at path root:solids:good_error_handling' in str(pe)


def test_multilevel_good_error_handling_config_solids_name_solids():
    @solid(config_field=Field(Nullable(Int)))
    def good_error_handling(_info):
        pass

    pipeline_def = PipelineDefinition(
        name='multilevel_good_error_handling', solids=[good_error_handling]
    )

    execute_pipeline(
        pipeline_def, environment={'solids': {'good_error_handling': {'config': None}}}
    )



# pylint: disable=W0622,W0614,W0401
import pytest

from dagster import (
    DagsterInvariantViolationError,
    DependencyDefinition,
    Field,
    InputDefinition,
    MultipleResults,
    OutputDefinition,
    PipelineDefinition,
    Result,
    execute_pipeline,
    solid,
    String,
    Int,
)


@solid(
    outputs=[
        OutputDefinition(dagster_type=Int, name='out_one'),
        OutputDefinition(dagster_type=Int, name='out_two'),
    ]
)
def yield_outputs(_info):
    yield Result(23, 'out_one')
    yield Result(45, 'out_two')


@solid(
    outputs=[
        OutputDefinition(dagster_type=Int, name='out_one'),
        OutputDefinition(dagster_type=Int, name='out_two'),
    ]
)
def return_dict_results(_info):
    return MultipleResults.from_dict({'out_one': 23, 'out_two': 45})


@solid(
    config_field=Field(
        String, description='Should be either out_one or out_two'
    ),
    outputs=[
        OutputDefinition(dagster_type=Int, name='out_one'),
        OutputDefinition(dagster_type=Int, name='out_two'),
    ],
)
def conditional(info):
    if info.config == 'out_one':
        yield Result(23, 'out_one')
    elif info.config == 'out_two':
        yield Result(45, 'out_two')
    else:
        raise Exception('invalid config')


@solid(inputs=[InputDefinition('num', dagster_type=Int)])
def log_num(info, num):
    info.context.info('num {num}'.format(num=num))
    return num


@solid(inputs=[InputDefinition('num', dagster_type=Int)])
def log_num_squared(info, num):
    info.context.info(
        'num_squared {num_squared}'.format(num_squared=num * num)
    )
    return num * num


def define_part_eleven_step_one_pipeline():
    return PipelineDefinition(
        name='part_eleven_step_one_pipeline',
        solids=[return_dict_results, log_num, log_num_squared],
        dependencies={
            'log_num': {
                'num': DependencyDefinition('return_dict_results', 'out_one')
            },
            'log_num_squared': {
                'num': DependencyDefinition('return_dict_results', 'out_two')
            },
        },
    )


def define_part_eleven_step_two_pipeline():
    return PipelineDefinition(
        name='part_eleven_step_two_pipeline',
        solids=[yield_outputs, log_num, log_num_squared],
        dependencies={
            'log_num': {
                'num': DependencyDefinition('yield_outputs', 'out_one')
            },
            'log_num_squared': {
                'num': DependencyDefinition('yield_outputs', 'out_two')
            },
        },
    )


def define_part_eleven_step_three_pipeline():
    return PipelineDefinition(
        name='part_eleven_step_three_pipeline',
        solids=[conditional, log_num, log_num_squared],
        dependencies={
            'log_num': {'num': DependencyDefinition('conditional', 'out_one')},
            'log_num_squared': {
                'num': DependencyDefinition('conditional', 'out_two')
            },
        },
    )


def test_intro_tutorial_part_eleven_step_one():
    result = execute_pipeline(define_part_eleven_step_one_pipeline())

    assert result.success
    assert (
        result.result_for_solid('return_dict_results').transformed_value(
            'out_one'
        )
        == 23
    )
    assert (
        result.result_for_solid('return_dict_results').transformed_value(
            'out_two'
        )
        == 45
    )
    assert result.result_for_solid('log_num').transformed_value() == 23
    assert (
        result.result_for_solid('log_num_squared').transformed_value()
        == 45 * 45
    )


def test_intro_tutorial_part_eleven_step_two():
    result = execute_pipeline(define_part_eleven_step_two_pipeline())

    assert result.success
    assert (
        result.result_for_solid('yield_outputs').transformed_value('out_one')
        == 23
    )
    assert (
        result.result_for_solid('yield_outputs').transformed_value('out_two')
        == 45
    )
    assert result.result_for_solid('log_num').transformed_value() == 23
    assert (
        result.result_for_solid('log_num_squared').transformed_value()
        == 45 * 45
    )


def test_intro_tutorial_part_eleven_step_three():
    result = execute_pipeline(
        define_part_eleven_step_three_pipeline(),
        {'solids': {'conditional': {'config': 'out_two'}}},
    )

    # successful things
    assert result.success
    assert (
        result.result_for_solid('conditional').transformed_value('out_two')
        == 45
    )
    assert (
        result.result_for_solid('log_num_squared').transformed_value()
        == 45 * 45
    )

    # unsuccessful things
    with pytest.raises(DagsterInvariantViolationError):
        assert (
            result.result_for_solid('conditional').transformed_value('out_one')
            == 45
        )


if __name__ == '__main__':
    execute_pipeline(
        define_part_eleven_step_three_pipeline(),
        {'solids': {'conditional': {'config': 'out_two'}}},
    )

    # execute_pipeline(define_part_eleven_step_two())



'''
Naming conventions:

For public functions:

execute_*

These represent functions which do purely in-memory compute. They will evaluate expectations
the core transform, and exercise all logging and metrics tracking (outside of outputs), but they
will not invoke *any* outputs (and their APIs don't allow the user to).


'''

# too many lines
# pylint: disable=C0302

from collections import defaultdict, namedtuple
from contextlib import contextmanager
import itertools
import inspect
import uuid

from contextlib2 import ExitStack
import six

from dagster import check

from .definitions import (
    DEFAULT_OUTPUT,
    ContextCreationExecutionInfo,
    DependencyDefinition,
    PipelineDefinition,
    Solid,
    SolidInstance,
    solids_in_topological_order,
)

from .execution_context import ExecutionContext, ReentrantInfo, RuntimeExecutionContext

from .errors import DagsterInvariantViolationError, DagsterUserCodeExecutionError

from .types.evaluator import EvaluationError, evaluate_config_value, friendly_string_for_error

from .events import construct_event_logger

from .execution_plan.create import create_execution_plan_core, create_subplan

from .execution_plan.objects import (
    ExecutionPlan,
    ExecutionPlanInfo,
    ExecutionPlanSubsetInfo,
    StepResult,
    StepTag,
)

from .execution_plan.simple_engine import execute_plan_core

from .system_config.objects import EnvironmentConfig
from .system_config.types import construct_environment_config


class PipelineExecutionResult(object):
    '''Result of execution of the whole pipeline. Returned eg by :py:function:`execute_pipeline`.

    Attributes:
        pipeline (PipelineDefinition): Pipeline that was executed
        context (ExecutionContext): ExecutionContext of that particular Pipeline run.
        result_list (list[SolidExecutionResult]): List of results for each pipeline solid.
    '''

    def __init__(self, pipeline, context, result_list):
        self.pipeline = check.inst_param(pipeline, 'pipeline', PipelineDefinition)
        self.context = check.inst_param(context, 'context', RuntimeExecutionContext)
        self.result_list = check.list_param(
            result_list, 'result_list', of_type=SolidExecutionResult
        )
        self.run_id = context.run_id

    @property
    def success(self):
        '''Whether the pipeline execution was successful at all steps'''
        return all([result.success for result in self.result_list])

    def result_for_solid(self, name):
        '''Get a :py:class:`SolidExecutionResult` for a given solid name.

        Returns:
          SolidExecutionResult
        '''
        check.str_param(name, 'name')

        if not self.pipeline.has_solid(name):
            raise DagsterInvariantViolationError(
                'Try to get result for solid {name} in {pipeline}. No such solid.'.format(
                    name=name, pipeline=self.pipeline.display_name
                )
            )

        for result in self.result_list:
            if result.solid.name == name:
                return result

        raise DagsterInvariantViolationError(
            'Did not find result for solid {name} in pipeline execution result'.format(name=name)
        )


class SolidExecutionResult(object):
    '''Execution result for one solid of the pipeline.

    Attributes:
      context (ExecutionContext): ExecutionContext of that particular Pipeline run.
      solid (SolidDefinition): Solid for which this result is
    '''

    def __init__(self, context, solid, step_results_by_tag):
        self.context = check.inst_param(context, 'context', RuntimeExecutionContext)
        self.solid = check.inst_param(solid, 'solid', Solid)
        self.step_results_by_tag = check.dict_param(
            step_results_by_tag, 'step_results_by_tag', key_type=StepTag, value_type=list
        )

    @property
    def transforms(self):
        return self.step_results_by_tag.get(StepTag.TRANSFORM, [])

    @property
    def input_expectations(self):
        return self.step_results_by_tag.get(StepTag.INPUT_EXPECTATION, [])

    @property
    def output_expectations(self):
        return self.step_results_by_tag.get(StepTag.OUTPUT_EXPECTATION, [])

    @staticmethod
    def from_results(context, results):
        check.inst_param(context, 'context', RuntimeExecutionContext)
        results = check.list_param(results, 'results', StepResult)
        if results:
            step_results_by_tag = defaultdict(list)

            solid = None
            for result in results:
                if solid is None:
                    solid = result.step.solid
                check.invariant(result.step.solid is solid, 'Must all be from same solid')

            for result in results:
                step_results_by_tag[result.tag].append(result)

            return SolidExecutionResult(
                context=context,
                solid=results[0].step.solid,
                step_results_by_tag=dict(step_results_by_tag),
            )
        else:
            check.failed("Cannot create SolidExecutionResult from empty list")

    @property
    def success(self):
        '''Whether the solid execution was successful'''
        return all(
            [
                result.success
                for result in itertools.chain(
                    self.input_expectations, self.output_expectations, self.transforms
                )
            ]
        )

    @property
    def transformed_values(self):
        '''Return dictionary of transformed results, with keys being output names.
        Returns None if execution result isn't a success.'''
        if self.success and self.transforms:
            return {
                result.success_data.output_name: result.success_data.value
                for result in self.transforms
            }
        else:
            return None

    def transformed_value(self, output_name=DEFAULT_OUTPUT):
        '''Returns transformed value either for DEFAULT_OUTPUT or for the output
        given as output_name. Returns None if execution result isn't a success'''
        check.str_param(output_name, 'output_name')

        if not self.solid.definition.has_output(output_name):
            raise DagsterInvariantViolationError(
                '{output_name} not defined in solid {solid}'.format(
                    output_name=output_name, solid=self.solid.name
                )
            )

        if self.success:
            for result in self.transforms:
                if result.success_data.output_name == output_name:
                    return result.success_data.value
            raise DagsterInvariantViolationError(
                (
                    'Did not find result {output_name} in solid {self.solid.name} '
                    'execution result'
                ).format(output_name=output_name, self=self)
            )
        else:
            return None

    def reraise_user_error(self):
        if not self.success:
            for result in itertools.chain(
                self.input_expectations, self.output_expectations, self.transforms
            ):
                if not result.success:
                    if isinstance(result.failure_data.dagster_error, DagsterUserCodeExecutionError):
                        six.reraise(*result.failure_data.dagster_error.original_exc_info)
                    else:
                        raise result.failure_data.dagster_error

    @property
    def dagster_error(self):
        '''Returns exception that happened during this solid's execution, if any'''
        for result in itertools.chain(
            self.input_expectations, self.output_expectations, self.transforms
        ):
            if not result.success:
                return result.failure_data.dagster_error


def create_execution_plan(pipeline, env_config=None):
    check.inst_param(pipeline, 'pipeline', PipelineDefinition)
    check.opt_dict_param(env_config, 'env_config', key_type=str)

    typed_environment = create_typed_environment(pipeline, env_config)
    return create_execution_plan_with_typed_environment(pipeline, typed_environment)


def create_execution_plan_with_typed_environment(pipeline, typed_environment):
    check.inst_param(pipeline, 'pipeline', PipelineDefinition)
    check.inst_param(typed_environment, 'environment', EnvironmentConfig)

    with yield_context(pipeline, typed_environment) as context:
        return create_execution_plan_core(ExecutionPlanInfo(context, pipeline, typed_environment))


def get_run_id(reentrant_info):
    check.opt_inst_param(reentrant_info, 'reentrant_info', ReentrantInfo)
    if reentrant_info and reentrant_info.run_id:
        return reentrant_info.run_id
    else:
        return str(uuid.uuid4())


def merge_two_dicts(left, right):
    result = left.copy()
    result.update(right)
    return result


def get_context_stack(user_context_params, reentrant_info):
    check.inst(user_context_params, ExecutionContext)
    check.opt_inst_param(reentrant_info, 'reentrant_info', ReentrantInfo)

    if reentrant_info and reentrant_info.context_stack:
        user_keys = set(user_context_params.context_stack.keys())
        reentrant_keys = set(reentrant_info.context_stack.keys())
        if not user_keys.isdisjoint(reentrant_keys):
            raise DagsterInvariantViolationError(
                (
                    'You have specified re-entrant keys and user-defined keys '
                    'that overlap. User keys: {user_keys}. Reentrant keys: '
                    '{reentrant_keys}.'
                ).format(user_keys=user_keys, reentrant_keys=reentrant_keys)
            )

        return merge_two_dicts(user_context_params.context_stack, reentrant_info.context_stack)
    else:
        return user_context_params.context_stack


ResourceCreationInfo = namedtuple('ResourceCreationInfo', 'config run_id')


def _ensure_gen(thing_or_gen):
    if not inspect.isgenerator(thing_or_gen):

        def _gen_thing():
            yield thing_or_gen

        return _gen_thing()

    return thing_or_gen


@contextmanager
def with_maybe_gen(thing_or_gen):
    gen = _ensure_gen(thing_or_gen)

    try:
        thing = next(gen)
    except StopIteration:
        check.failed('Must yield one item. You did not yield anything.')

    yield thing

    stopped = False

    try:
        next(gen)
    except StopIteration:
        stopped = True

    check.invariant(stopped, 'Must yield one item. Yielded more than one item')


@contextmanager
def yield_context(pipeline, environment, reentrant_info=None):
    check.inst_param(pipeline, 'pipeline', PipelineDefinition)
    check.inst_param(environment, 'environment', EnvironmentConfig)
    check.opt_inst_param(reentrant_info, 'reentrant_info', ReentrantInfo)

    context_definition = pipeline.context_definitions[environment.context.name]

    run_id = get_run_id(reentrant_info)

    ec_or_gen = context_definition.context_fn(
        ContextCreationExecutionInfo(
            config=environment.context.config, pipeline_def=pipeline, run_id=run_id
        )
    )

    with with_maybe_gen(ec_or_gen) as execution_context:
        check.inst(execution_context, ExecutionContext)

        with _create_resources(
            pipeline, context_definition, environment, execution_context, run_id
        ) as resources:
            loggers = _create_loggers(reentrant_info, execution_context)

            yield RuntimeExecutionContext(
                run_id=get_run_id(reentrant_info),
                loggers=loggers,
                resources=resources,
                context_stack=get_context_stack(execution_context, reentrant_info),
            )


def _create_loggers(reentrant_info, execution_context):
    if reentrant_info and reentrant_info.event_callback:
        return execution_context.loggers + [construct_event_logger(reentrant_info.event_callback)]
    else:
        return execution_context.loggers


@contextmanager
def _create_resources(pipeline_def, context_def, environment, execution_context, run_id):
    if not context_def.resources:
        yield execution_context.resources
        return

    resources = {}
    check.invariant(
        not execution_context.resources,
        (
            'If resources explicitly specified on context definition, the context '
            'creation function should not return resources as a property of the '
            'ExecutionContext.'
        ),
    )

    # See https://bit.ly/2zIXyqw
    # The "ExitStack" allows one to stack up N context managers and then yield
    # something. We do this so that resources can cleanup after themselves. We
    # can potentially have many resources so we need to use this abstraction.
    with ExitStack() as stack:
        for resource_name in context_def.resources.keys():
            resource_obj_or_gen = get_resource_or_gen(
                context_def, resource_name, environment, run_id
            )

            resource_obj = stack.enter_context(with_maybe_gen(resource_obj_or_gen))

            resources[resource_name] = resource_obj

        context_name = environment.context.name

        resources_type = pipeline_def.context_definitions[context_name].resources_type
        yield resources_type(**resources)


def get_resource_or_gen(context_definition, resource_name, environment, run_id):
    resource_def = context_definition.resources[resource_name]
    # Need to do default values
    resource_config = environment.context.resources.get(resource_name, {}).get('config')
    return resource_def.resource_fn(ResourceCreationInfo(resource_config, run_id))


def _do_iterate_pipeline(pipeline, context, typed_environment, throw_on_error=True):
    check.inst(context, RuntimeExecutionContext)
    pipeline_success = True
    with context.value('pipeline', pipeline.display_name):
        context.events.pipeline_start()

        execution_plan = create_execution_plan_core(
            ExecutionPlanInfo(context, pipeline, typed_environment)
        )

        steps = list(execution_plan.topological_steps())

        if not steps:
            context.debug(
                'Pipeline {pipeline} has no nodes and no execution will happen'.format(
                    pipeline=pipeline.display_name
                )
            )
            context.events.pipeline_success()
            return

        context.debug(
            'About to execute the compute node graph in the following order {order}'.format(
                order=[step.key for step in steps]
            )
        )

        check.invariant(len(steps[0].step_inputs) == 0)

        for solid_result in _process_step_results(
            context, pipeline, execute_plan_core(context, execution_plan)
        ):
            if throw_on_error and not solid_result.success:
                solid_result.reraise_user_error()

            if not solid_result.success:
                pipeline_success = False
            yield solid_result

        if pipeline_success:
            context.events.pipeline_success()
        else:
            context.events.pipeline_failure()


def execute_pipeline_iterator(
    pipeline, environment=None, throw_on_error=True, reentrant_info=None, solid_subset=None
):
    '''Returns iterator that yields :py:class:`SolidExecutionResult` for each
    solid executed in the pipeline.

    This is intended to allow the caller to do things between each executed
    node. For the 'synchronous' API, see :py:function:`execute_pipeline`.

    Parameters:
      pipeline (PipelineDefinition): pipeline to run
      execution (ExecutionContext): execution context of the run
    '''
    check.inst_param(pipeline, 'pipeline', PipelineDefinition)
    check.opt_dict_param(environment, 'environment')
    check.bool_param(throw_on_error, 'throw_on_error')
    check.opt_inst_param(reentrant_info, 'reentrant_info', ReentrantInfo)
    check.opt_list_param(solid_subset, 'solid_subset', of_type=str)

    pipeline_to_execute = get_subset_pipeline(pipeline, solid_subset)
    typed_environment = create_typed_environment(pipeline_to_execute, environment)

    with yield_context(pipeline_to_execute, typed_environment, reentrant_info) as context:
        for solid_result in _do_iterate_pipeline(
            pipeline_to_execute, context, typed_environment, throw_on_error
        ):
            yield solid_result


def _process_step_results(context, pipeline, step_results):
    check.inst_param(context, 'context', RuntimeExecutionContext)
    check.inst_param(pipeline, 'pipeline', PipelineDefinition)

    step_results_by_solid_name = defaultdict(list)
    for step_result in step_results:
        step_results_by_solid_name[step_result.step.solid.name].append(step_result)

    for topo_solid in solids_in_topological_order(pipeline):
        if topo_solid.name in step_results_by_solid_name:
            yield SolidExecutionResult.from_results(
                context, step_results_by_solid_name[topo_solid.name]
            )


class PipelineConfigEvaluationError(Exception):
    def __init__(self, pipeline, errors, config_value, *args, **kwargs):
        self.pipeline = check.inst_param(pipeline, 'pipeline', PipelineDefinition)
        self.errors = check.list_param(errors, 'errors', of_type=EvaluationError)
        self.config_value = config_value

        error_msg = 'Pipeline "{pipeline}" config errors:'.format(pipeline=pipeline.name)

        error_messages = []

        for i_error, error in enumerate(self.errors):
            error_message = friendly_string_for_error(error)
            error_messages.append(error_message)
            error_msg += '\n    Error {i_error}: {error_message}'.format(
                i_error=i_error + 1, error_message=error_message
            )

        self.message = error_msg
        self.error_messages = error_messages

        super(PipelineConfigEvaluationError, self).__init__(error_msg, *args, **kwargs)


def execute_plan(pipeline, execution_plan, environment=None, subset_info=None, reentrant_info=None):
    check.inst_param(pipeline, 'pipeline', PipelineDefinition)
    check.inst_param(execution_plan, 'execution_plan', ExecutionPlan)
    check.opt_dict_param(environment, 'environment')
    check.opt_inst_param(subset_info, 'subset_info', ExecutionPlanSubsetInfo)
    check.opt_inst_param(reentrant_info, 'reentrant_info', ReentrantInfo)

    typed_environment = create_typed_environment(pipeline, environment)

    with yield_context(pipeline, typed_environment, reentrant_info) as context:
        plan_to_execute = (
            create_subplan(
                ExecutionPlanInfo(
                    context=context, pipeline=pipeline, environment=typed_environment
                ),
                execution_plan,
                subset_info,
            )
            if subset_info
            else execution_plan
        )
        return list(execute_plan_core(context, plan_to_execute))


def execute_pipeline(
    pipeline, environment=None, throw_on_error=True, reentrant_info=None, solid_subset=None
):
    '''
    "Synchronous" version of :py:function:`execute_pipeline_iterator`.

    Note: throw_on_error is very useful in testing contexts when not testing for error conditions

    Parameters:
      pipeline (PipelineDefinition): Pipeline to run
      environment (dict): The enviroment that parameterizes this run
      throw_on_error (bool):
        throw_on_error makes the function throw when an error is encoutered rather than returning
        the py:class:`SolidExecutionResult` in an error-state.


    Returns:
      PipelineExecutionResult
    '''

    check.inst_param(pipeline, 'pipeline', PipelineDefinition)
    check.opt_dict_param(environment, 'environment')
    check.bool_param(throw_on_error, 'throw_on_error')
    check.opt_inst_param(reentrant_info, 'reentrant_info', ReentrantInfo)
    check.opt_list_param(solid_subset, 'solid_subset', of_type=str)

    pipeline_to_execute = get_subset_pipeline(pipeline, solid_subset)
    typed_environment = create_typed_environment(pipeline_to_execute, environment)
    return execute_reentrant_pipeline(
        pipeline_to_execute, typed_environment, throw_on_error, reentrant_info
    )


def _dep_key_of(solid):
    return SolidInstance(solid.definition.name, solid.name)


def build_sub_pipeline(pipeline_def, solid_names):
    '''
    Build a pipeline which is a subset of another pipeline.
    Only includes the solids which are in solid_names.
    '''

    check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)
    check.list_param(solid_names, 'solid_names', of_type=str)

    solid_name_set = set(solid_names)
    solids = list(map(pipeline_def.solid_named, solid_names))
    deps = {_dep_key_of(solid): {} for solid in solids}

    def _out_handle_of_inp(input_handle):
        if pipeline_def.dependency_structure.has_dep(input_handle):
            output_handle = pipeline_def.dependency_structure.get_dep(input_handle)
            if output_handle.solid.name in solid_name_set:
                return output_handle
        return None

    for solid in solids:
        for input_handle in solid.input_handles():
            output_handle = _out_handle_of_inp(input_handle)
            if output_handle:
                deps[_dep_key_of(solid)][input_handle.input_def.name] = DependencyDefinition(
                    solid=output_handle.solid.name, output=output_handle.output_def.name
                )

    return PipelineDefinition(
        name=pipeline_def.name,
        solids=list(set([solid.definition for solid in solids])),
        context_definitions=pipeline_def.context_definitions,
        dependencies=deps,
    )


def execute_reentrant_pipeline(pipeline, typed_environment, throw_on_error, reentrant_info):
    check.inst_param(pipeline, 'pipeline', PipelineDefinition)
    check.inst_param(typed_environment, 'typed_environment', EnvironmentConfig)
    check.opt_inst_param(reentrant_info, 'reentrant_info', ReentrantInfo)

    with yield_context(pipeline, typed_environment, reentrant_info) as context:
        return PipelineExecutionResult(
            pipeline,
            context,
            list(_do_iterate_pipeline(pipeline, context, typed_environment, throw_on_error)),
        )


def get_subset_pipeline(pipeline, solid_subset):
    check.inst_param(pipeline, 'pipeline', PipelineDefinition)
    check.opt_list_param(solid_subset, 'solid_subset', of_type=str)
    return pipeline if solid_subset is None else build_sub_pipeline(pipeline, solid_subset)


def create_typed_environment(pipeline, environment=None):
    check.inst_param(pipeline, 'pipeline', PipelineDefinition)
    check.opt_dict_param(environment, 'environment')

    result = evaluate_config_value(pipeline.environment_type, environment)

    if not result.success:
        raise PipelineConfigEvaluationError(pipeline, result.errors, environment)

    return construct_environment_config(result.value)



from contextlib import contextmanager
import sys

from future.utils import raise_from

from dagster import check

from dagster.utils.logging import get_formatted_stack_trace

from dagster.utils.timing import time_execution_scope

from dagster.core.definitions import Result

from dagster.core.errors import (
    DagsterError,
    DagsterInvariantViolationError,
    DagsterRuntimeCoercionError,
    DagsterUserCodeExecutionError,
    DagsterTypeError,
)

from dagster.core.execution_context import RuntimeExecutionContext

from .objects import (
    ExecutionPlan,
    ExecutionStep,
    StepResult,
    StepOutputHandle,
    StepSuccessData,
    StepFailureData,
)


def _all_inputs_covered(step, results):
    for step_input in step.step_inputs:
        if step_input.prev_output_handle not in results:
            return False
    return True


def execute_plan_core(context, execution_plan):
    check.inst_param(context, 'context', RuntimeExecutionContext)
    check.inst_param(execution_plan, 'execution_plan', ExecutionPlan)
    steps = list(execution_plan.topological_steps())

    intermediate_results = {}
    context.debug(
        'Entering execute_steps loop. Order: {order}'.format(order=[step.key for step in steps])
    )

    for step in steps:
        if not _all_inputs_covered(step, intermediate_results):
            result_keys = set(intermediate_results.keys())
            expected_outputs = [ni.prev_output_handle for ni in step.step_inputs]

            context.debug(
                'Not all inputs covered for {step}. Not executing.'.format(step=step.key)
                + '\nKeys in result: {result_keys}.'.format(result_keys=result_keys)
                + '\nOutputs need for inputs {expected_outputs}'.format(
                    expected_outputs=expected_outputs
                )
            )
            continue

        input_values = {}
        for step_input in step.step_inputs:
            prev_output_handle = step_input.prev_output_handle
            input_value = intermediate_results[prev_output_handle].success_data.value
            input_values[step_input.name] = input_value

        for result in execute_step(step, context, input_values):
            check.invariant(isinstance(result, StepResult))
            yield result
            if result.success:
                output_handle = StepOutputHandle(step, result.success_data.output_name)
                intermediate_results[output_handle] = result


def execute_step(step, context, inputs):
    check.inst_param(step, 'step', ExecutionStep)
    check.inst_param(context, 'context', RuntimeExecutionContext)
    check.dict_param(inputs, 'inputs', key_type=str)

    try:
        for step_result in _execute_steps_core_loop(step, context, inputs):
            context.info(
                'Step {step} emitted {value} for output {output}'.format(
                    step=step.key,
                    value=repr(step_result.success_data.value),
                    output=step_result.success_data.output_name,
                )
            )
            yield step_result
    except DagsterError as dagster_error:
        context.error(str(dagster_error))
        yield StepResult.failure_result(
            step=step, tag=step.tag, failure_data=StepFailureData(dagster_error=dagster_error)
        )
        return


def _error_check_results(step, results):
    seen_outputs = set()
    for result in results:
        if not step.has_step_output(result.output_name):
            output_names = list(
                [output_def.name for output_def in step.solid.definition.output_defs]
            )
            raise DagsterInvariantViolationError(
                '''Core transform for {step.solid.name} returned an output
                {result.output_name} that does not exist. The available
                outputs are {output_names}'''.format(
                    step=step, result=result, output_names=output_names
                )
            )

        if result.output_name in seen_outputs:
            raise DagsterInvariantViolationError(
                '''Core transform for {step.solid.name} returned an output
                {result.output_name} multiple times'''.format(
                    step=step, result=result
                )
            )

        seen_outputs.add(result.output_name)


def _execute_steps_core_loop(step, context, inputs):
    evaluated_inputs = {}
    # do runtime type checks of inputs versus step inputs
    for input_name, input_value in inputs.items():
        evaluated_inputs[input_name] = _get_evaluated_input(step, input_name, input_value)

    results = _compute_result_list(step, context, evaluated_inputs)

    _error_check_results(step, results)

    return [_create_step_result(step, result) for result in results]


def _create_step_result(step, result):
    check.inst_param(result, 'result', Result)

    step_output = step.step_output_named(result.output_name)

    try:
        coerced_value = step_output.runtime_type.coerce_runtime_value(result.value)
    except DagsterRuntimeCoercionError as e:
        raise DagsterInvariantViolationError(
            '''Solid {step.solid.name} output name {output_name} output {result.value}
            type failure: {error_msg}'''.format(
                step=step, result=result, error_msg=','.join(e.args), output_name=result.output_name
            )
        )

    return StepResult.success_result(
        step=step,
        tag=step.tag,
        success_data=StepSuccessData(output_name=result.output_name, value=coerced_value),
    )


def _get_evaluated_input(step, input_name, input_value):
    step_input = step.step_input_named(input_name)
    try:
        return step_input.runtime_type.coerce_runtime_value(input_value)
    except DagsterRuntimeCoercionError as evaluate_error:
        raise_from(
            DagsterTypeError(
                (
                    'Solid {step.solid.name} input {input_name} received value {input_value} '
                    'which does not pass the typecheck for Dagster type '
                    '{step_input.runtime_type.name}. Step {step.key}'
                ).format(
                    step=step, input_name=input_name, input_value=input_value, step_input=step_input
                )
            ),
            evaluate_error,
        )


def _compute_result_list(step, context, evaluated_inputs):
    error_str = 'Error occured during step {key}'.format(key=step.key)

    def_name = step.solid.definition.name

    with context.values({'solid': step.solid.name, 'solid_definition': def_name}):
        with _execution_step_error_boundary(context, step, error_str):
            gen = step.compute_fn(context, step, evaluated_inputs)

            if gen is None:
                check.invariant(not step.step_outputs)
                return

            results = list(gen)

        return results


@contextmanager
def _execution_step_error_boundary(context, step, msg, **kwargs):
    '''
    Wraps the execution of user-space code in an error boundary. This places a uniform
    policy around an user code invoked by the framework. This ensures that all user
    errors are wrapped in the SolidUserCodeExecutionError, and that the original stack
    trace of the user error is preserved, so that it can be reported without confusing
    framework code in the stack trace, if a tool author wishes to do so. This has
    been especially help in a notebooking context.
    '''
    check.inst_param(context, 'context', RuntimeExecutionContext)
    check.str_param(msg, 'msg')

    context.events.execution_plan_step_start(step.key)
    try:
        with time_execution_scope() as timer_result:
            yield

        context.events.execution_plan_step_success(step.key, timer_result.millis)
    except Exception as e:  # pylint: disable=W0703
        context.events.execution_plan_step_failure(step.key, sys.exc_info())

        stack_trace = get_formatted_stack_trace(e)
        context.error(str(e), stack_trace=stack_trace)

        if isinstance(e, DagsterError):
            raise e
        else:
            raise_from(
                DagsterUserCodeExecutionError(
                    msg.format(**kwargs), user_exception=e, original_exc_info=sys.exc_info()
                ),
                e,
            )



import pytest

from dagster import (
    DagsterInvariantViolationError,
    ExecutionContext,
    PipelineContextDefinition,
    PipelineDefinition,
    ReentrantInfo,
    execute_pipeline,
    solid,
)


def test_injected_run_id():
    run_id = 'kdjfkdjfkd'
    pipeline_def = PipelineDefinition(name='injected_run_id', solids=[])
    result = execute_pipeline(pipeline_def, reentrant_info=ReentrantInfo(run_id=run_id))

    assert result.success
    assert result.context.run_id == run_id


def test_injected_context_stack():
    called = {}

    @solid
    def check_context_stack(info):
        assert info.context.get_context_value('foo') == 'bar'
        called['yup'] = True

    pipeline_def = PipelineDefinition(name='injected_run_id', solids=[check_context_stack])
    result = execute_pipeline(
        pipeline_def, reentrant_info=ReentrantInfo(context_stack={'foo': 'bar'})
    )

    assert result.success
    assert called['yup']


def test_user_injected_context_stack():
    called = {}

    @solid
    def check_context_stack(info):
        assert info.context.get_context_value('foo') == 'bar'
        assert info.context.get_context_value('quux') == 'baaz'
        called['yup'] = True

    def _create_context(_info):
        return ExecutionContext(context_stack={'quux': 'baaz'})

    pipeline_def = PipelineDefinition(
        name='injected_run_id',
        solids=[check_context_stack],
        context_definitions={'default': PipelineContextDefinition(context_fn=_create_context)},
    )

    result = execute_pipeline(
        pipeline_def, reentrant_info=ReentrantInfo(context_stack={'foo': 'bar'})
    )

    assert result.success
    assert called['yup']


def test_user_injected_context_stack_collision():
    called = {}

    @solid
    def check_context_stack(info):
        assert info.context.get_context_value('foo') == 'bar'
        assert info.context.get_context_value('quux') == 'baaz'
        called['yup'] = True

    def _create_context(_info):
        return ExecutionContext(context_stack={'foo': 'baaz'})

    pipeline_def = PipelineDefinition(
        name='injected_run_id',
        solids=[check_context_stack],
        context_definitions={'default': PipelineContextDefinition(context_fn=_create_context)},
    )

    with pytest.raises(DagsterInvariantViolationError, match='You have specified'):
        execute_pipeline(pipeline_def, reentrant_info=ReentrantInfo(context_stack={'foo': 'bar'}))



from collections import defaultdict
from dagster import check

from dagster.core.types import config
from dagster.core.types.iterate_types import iterate_config_types
from dagster.core.errors import DagsterInvalidDefinitionError

from .context import PipelineContextDefinition

from .dependency import DependencyStructure, SolidInstance, Solid

from .solid import SolidDefinition


class SolidAliasMapper:
    def __init__(self, dependencies_dict):
        aliased_dependencies_dict = {}
        solid_uses = defaultdict(set)
        alias_lookup = {}

        for solid_key, input_dep_dict in dependencies_dict.items():
            if not isinstance(solid_key, SolidInstance):
                solid_key = SolidInstance(solid_key)

            if solid_key.alias:
                key = solid_key.name
                alias = solid_key.alias
            else:
                key = solid_key.name
                alias = solid_key.name

            solid_uses[key].add(alias)
            aliased_dependencies_dict[alias] = input_dep_dict
            alias_lookup[alias] = key

            for dependency in input_dep_dict.values():
                solid_uses[dependency.solid].add(dependency.solid)

        self.solid_uses = solid_uses
        self.aliased_dependencies_dict = aliased_dependencies_dict
        self.alias_lookup = alias_lookup

    def get_uses_of_solid(self, solid_def_name):
        return self.solid_uses.get(solid_def_name)


def create_execution_structure(solids, dependencies_dict):
    mapper = SolidAliasMapper(dependencies_dict)

    pipeline_solids = []
    for solid_def in solids:
        if isinstance(solid_def, SolidDefinition):
            uses_of_solid = mapper.get_uses_of_solid(solid_def.name) or set([solid_def.name])

            for alias in uses_of_solid:
                pipeline_solids.append(Solid(name=alias, definition=solid_def))

        elif callable(solid_def):
            raise DagsterInvalidDefinitionError(
                '''You have passed a lambda or function {func} into a pipeline that is
                not a solid. You have likely forgetten to annotate this function with
                an @solid or @lambda_solid decorator located in dagster.core.decorators
                '''.format(
                    func=solid_def.__name__
                )
            )
        else:
            raise DagsterInvalidDefinitionError(
                'Invalid item in solid list: {item}'.format(item=repr(solid_def))
            )

    pipeline_solid_dict = {ps.name: ps for ps in pipeline_solids}

    _validate_dependencies(
        mapper.aliased_dependencies_dict, pipeline_solid_dict, mapper.alias_lookup
    )

    dependency_structure = DependencyStructure.from_definitions(
        pipeline_solid_dict, mapper.aliased_dependencies_dict
    )

    return dependency_structure, pipeline_solid_dict


def _validate_dependencies(dependencies, solid_dict, alias_lookup):
    for from_solid, dep_by_input in dependencies.items():
        for from_input, dep in dep_by_input.items():
            if from_solid == dep.solid:
                raise DagsterInvalidDefinitionError(
                    'Circular reference detected in solid {from_solid} input {from_input}.'.format(
                        from_solid=from_solid, from_input=from_input
                    )
                )

            if not from_solid in solid_dict:
                aliased_solid = alias_lookup.get(from_solid)
                if aliased_solid == from_solid:
                    raise DagsterInvalidDefinitionError(
                        'Solid {solid} in dependency dictionary not found in solid list'.format(
                            solid=from_solid
                        )
                    )
                else:
                    raise DagsterInvalidDefinitionError(
                        (
                            'Solid {aliased_solid} (aliased by {from_solid} in dependency '
                            'dictionary) not found in solid list'
                        ).format(aliased_solid=aliased_solid, from_solid=from_solid)
                    )
            if not solid_dict[from_solid].definition.has_input(from_input):
                input_list = [
                    input_def.name for input_def in solid_dict[from_solid].definition.input_defs
                ]
                raise DagsterInvalidDefinitionError(
                    'Solid "{from_solid}" does not have input "{from_input}". '.format(
                        from_solid=from_solid, from_input=from_input
                    )
                    + 'Input list: {input_list}'.format(input_list=input_list)
                )

            if not dep.solid in solid_dict:
                raise DagsterInvalidDefinitionError(
                    'Solid {dep.solid} in DependencyDefinition not found in solid list'.format(
                        dep=dep
                    )
                )

            if not solid_dict[dep.solid].definition.has_output(dep.output):
                raise DagsterInvalidDefinitionError(
                    'Solid {dep.solid} does not have output {dep.output}'.format(dep=dep)
                )


def iterate_solid_def_types(solid_def):
    if solid_def.config_field:
        for runtime_type in iterate_config_types(solid_def.config_field.config_type):
            yield runtime_type


def _gather_all_config_types(solid_defs, context_definitions, environment_type):
    check.list_param(solid_defs, 'solid_defs', SolidDefinition)
    check.dict_param(
        context_definitions,
        'context_definitions',
        key_type=str,
        value_type=PipelineContextDefinition,
    )

    check.inst_param(environment_type, 'environment_type', config.ConfigType)

    for solid_def in solid_defs:
        for runtime_type in iterate_solid_def_types(solid_def):
            yield runtime_type

    for context_definition in context_definitions.values():
        if context_definition.config_field:
            context_config_type = context_definition.config_field.config_type
            for runtime_type in iterate_config_types(context_config_type):
                yield runtime_type

    for runtime_type in iterate_config_types(environment_type):
        yield runtime_type


def construct_runtime_type_dictionary(solid_defs):
    type_dict = {}
    for solid_def in solid_defs:
        for input_def in solid_def.input_defs:
            type_dict[input_def.runtime_type.name] = input_def.runtime_type

        for output_def in solid_def.output_defs:
            type_dict[output_def.runtime_type.name] = output_def.runtime_type

    return type_dict


def _gather_all_schemas(solid_defs):
    runtime_types = construct_runtime_type_dictionary(solid_defs)
    for rtt in runtime_types.values():
        if rtt.input_schema:
            for ct in iterate_config_types(rtt.input_schema.schema_type):
                yield ct
        if rtt.output_schema:
            for ct in iterate_config_types(rtt.output_schema.schema_type):
                yield ct


def construct_config_type_dictionary(solid_defs, context_definitions, environment_type):
    check.list_param(solid_defs, 'solid_defs', SolidDefinition)
    check.dict_param(
        context_definitions,
        'context_definitions',
        key_type=str,
        value_type=PipelineContextDefinition,
    )
    check.inst_param(environment_type, 'environment_type', config.ConfigType)

    type_dict = {}
    all_types = list(
        _gather_all_config_types(solid_defs, context_definitions, environment_type)
    ) + list(_gather_all_schemas(solid_defs))
    for config_type in all_types:
        name = config_type.name
        if name in type_dict:
            if type(config_type) is not type(type_dict[name]):
                raise DagsterInvalidDefinitionError(
                    (
                        'Type names must be unique. You have construct two instances of types '
                        'with the same name {name} but have different instances. Instance one '
                        '{inst_one}. Instance two {inst_two}'
                    ).format(name=name, inst_one=type(config_type), inst_two=type(type_dict[name]))
                )
        else:
            type_dict[config_type.name] = config_type

    return type_dict



import inspect
import os
import re
import yaml

from dagster import check

from .yaml_utils import load_yaml_from_path, load_yaml_from_globs, load_yaml_from_glob_list


def script_relative_path(file_path):
    '''
    Useful for testing with local files. Use a path relative to where the
    test resides and this function will return the absolute path
    of that file. Otherwise it will be relative to script that
    ran the test
    '''
    # from http://bit.ly/2snyC6s

    check.str_param(file_path, 'file_path')
    scriptdir = inspect.stack()[1][1]
    return os.path.join(os.path.dirname(os.path.abspath(scriptdir)), file_path)


# Adapted from https://github.com/okunishinishi/python-stringcase/blob/master/stringcase.py
def camelcase(string):
    check.str_param(string, 'string')

    string = re.sub(r'^[\-_\.]', '', str(string))
    if not string:
        return string
    return str(string[0]).upper() + re.sub(
        r'[\-_\.\s]([a-z])', lambda matched: str(matched.group(1)).upper(), string[1:]
    )



import click

from .pipeline import create_pipeline_cli


def create_dagster_cli():
    group = click.Group()
    group.add_command(create_pipeline_cli())
    return group


def main():
    cli = create_dagster_cli()
    cli(obj={})



# encoding: utf-8

from dagster import Field, PipelineDefinition, execute_pipeline, solid, types


@solid(
    config_field=Field(types.String, is_optional=True, default_value='en-us')
)
def configurable_hello(info):
    if len(info.config) >= 3 and info.config[:3] == 'haw':
        return 'Aloha honua!'
    elif len(info.config) >= 2 and info.config[:2] == 'cn':
        return '你好, 世界!'
    else:
        return 'Hello, world!'


def define_configurable_hello_pipeline():
    return PipelineDefinition(
        name='configurable_hello_pipeline', solids=[configurable_hello]
    )


def test_intro_tutorial_part_four():
    execute_pipeline(
        define_configurable_hello_pipeline(),
        {'solids': {'configurable_hello': {'config': 'cn'}}},
    )



from dagster import (
    InputDefinition,
    OutputDefinition,
    PipelineDefinition,
    execute_pipeline,
    lambda_solid,
    types,
)


@lambda_solid(inputs=[InputDefinition('word')])
def add_hello_to_word(word):
    return 'Hello, ' + word + '!'


def define_hello_inputs_pipeline():
    return PipelineDefinition(name='hello_inputs', solids=[add_hello_to_word])


def execute_with_another_world():
    return execute_pipeline(
        define_hello_inputs_pipeline(),
        # This entire dictionary is known as the 'environment'.
        # It has many sections.
        {
            # This is the 'solids' section
            'solids': {
                # Configuration for the add_hello_to_word solid
                'add_hello_to_word': {'inputs': {'word': 'Mars'}}
            }
        },
    )


@lambda_solid(
    inputs=[InputDefinition('word', types.String)],
    output=OutputDefinition(types.String),
)
def add_hello_to_word_typed(word):
    return 'Hello, ' + word + '!'


def define_hello_typed_inputs_pipeline():
    return PipelineDefinition(
        name='hello_typed_inputs', solids=[add_hello_to_word_typed]
    )



from abc import ABCMeta, abstractmethod
import pickle

import six


@six.add_metaclass(ABCMeta)
class MarshallingStrategy:
    @abstractmethod
    def marshal_value(self, value, to_file):
        pass

    @abstractmethod
    def unmarshal_value(self, from_file):
        pass


class PickleMarshallingStrategy(MarshallingStrategy):
    def marshal_value(self, value, to_file):
        with open(to_file, 'wb') as ff:
            pickle.dump(value, ff)

    def unmarshal_value(self, from_file):
        with open(from_file, 'rb') as ff:
            return pickle.load(ff)



from dagster import check

from dagster.core.definitions import (
    ExpectationDefinition,
    ExpectationExecutionInfo,
    InputDefinition,
    OutputDefinition,
    Result,
    Solid,
)

from dagster.core.errors import DagsterExpectationFailedError

from .objects import (
    ExecutionPlanInfo,
    ExecutionStep,
    ExecutionValueSubPlan,
    StepInput,
    StepOutput,
    StepOutputHandle,
    StepTag,
)

from .utility import create_joining_subplan

EXPECTATION_INPUT = 'expectation_input'
EXPECTATION_VALUE_OUTPUT = 'expectation_value'


def _create_expectation_lambda(solid, inout_def, expectation_def, internal_output_name):
    check.inst_param(solid, 'solid', Solid)
    check.inst_param(inout_def, 'inout_def', (InputDefinition, OutputDefinition))
    check.inst_param(expectation_def, 'expectations_def', ExpectationDefinition)
    check.str_param(internal_output_name, 'internal_output_name')

    def _do_expectation(context, step, inputs):
        with context.values(
            {inout_def.descriptive_key: inout_def.name, 'expectation': expectation_def.name}
        ):
            value = inputs[EXPECTATION_INPUT]
            info = ExpectationExecutionInfo(context, inout_def, solid, expectation_def)
            expt_result = expectation_def.expectation_fn(info, value)
            if expt_result.success:
                context.debug(
                    'Expectation {key} succeeded on {value}.'.format(key=step.key, value=value)
                )
                yield Result(output_name=internal_output_name, value=inputs[EXPECTATION_INPUT])
            else:
                context.debug(
                    'Expectation {key} failed on {value}.'.format(key=step.key, value=value)
                )
                raise DagsterExpectationFailedError(info, value)

    return _do_expectation


def create_expectations_subplan(solid, inout_def, prev_step_output_handle, tag):
    check.inst_param(solid, 'solid', Solid)
    check.inst_param(inout_def, 'inout_def', (InputDefinition, OutputDefinition))
    check.inst_param(prev_step_output_handle, 'prev_step_output_handle', StepOutputHandle)
    check.inst_param(tag, 'tag', StepTag)

    input_expect_steps = []
    for expectation_def in inout_def.expectations:
        expect_step = create_expectation_step(
            solid=solid,
            expectation_def=expectation_def,
            key='{solid}.{desc_key}.{inout_name}.expectation.{expectation_name}'.format(
                solid=solid.name,
                desc_key=inout_def.descriptive_key,
                inout_name=inout_def.name,
                expectation_name=expectation_def.name,
            ),
            tag=tag,
            prev_step_output_handle=prev_step_output_handle,
            inout_def=inout_def,
        )
        input_expect_steps.append(expect_step)

    return create_joining_subplan(
        solid,
        '{solid}.{desc_key}.{inout_name}.expectations.join'.format(
            solid=solid.name, desc_key=inout_def.descriptive_key, inout_name=inout_def.name
        ),
        input_expect_steps,
        EXPECTATION_VALUE_OUTPUT,
    )


def create_expectation_step(solid, expectation_def, key, tag, prev_step_output_handle, inout_def):

    check.inst_param(solid, 'solid', Solid)
    check.inst_param(expectation_def, 'input_expct_def', ExpectationDefinition)
    check.inst_param(prev_step_output_handle, 'prev_step_output_handle', StepOutputHandle)
    check.inst_param(inout_def, 'inout_def', (InputDefinition, OutputDefinition))

    value_type = inout_def.runtime_type

    return ExecutionStep(
        key=key,
        step_inputs=[
            StepInput(
                name=EXPECTATION_INPUT,
                runtime_type=value_type,
                prev_output_handle=prev_step_output_handle,
            )
        ],
        step_outputs=[StepOutput(name=EXPECTATION_VALUE_OUTPUT, runtime_type=value_type)],
        compute_fn=_create_expectation_lambda(
            solid, inout_def, expectation_def, EXPECTATION_VALUE_OUTPUT
        ),
        tag=tag,
        solid=solid,
    )


def decorate_with_expectations(execution_info, solid, transform_step, output_def):
    check.inst_param(execution_info, 'execution_info', ExecutionPlanInfo)
    check.inst_param(solid, 'solid', Solid)
    check.inst_param(transform_step, 'transform_step', ExecutionStep)
    check.inst_param(output_def, 'output_def', OutputDefinition)

    if execution_info.environment.expectations.evaluate and output_def.expectations:
        return create_expectations_subplan(
            solid,
            output_def,
            StepOutputHandle(transform_step, output_def.name),
            tag=StepTag.OUTPUT_EXPECTATION,
        )
    else:
        return ExecutionValueSubPlan.empty(StepOutputHandle(transform_step, output_def.name))



from collections import namedtuple

import six

from dagster import check

from .builtin_enum import BuiltinEnum

# from .dagster_type import check_dagster_type_param
from .wrapping import WrappingListType, WrappingNullableType


class ConfigTypeAttributes(
    namedtuple('_ConfigTypeAttributes', 'is_builtin is_system_config is_named')
):
    def __new__(cls, is_builtin=False, is_system_config=False, is_named=True):
        return super(ConfigTypeAttributes, cls).__new__(
            cls,
            is_builtin=check.bool_param(is_builtin, 'is_builtin'),
            is_system_config=check.bool_param(is_system_config, 'is_system_config'),
            is_named=check.bool_param(is_named, 'is_named'),
        )


DEFAULT_TYPE_ATTRIBUTES = ConfigTypeAttributes()


class ConfigType(object):
    def __init__(self, name=None, type_attributes=DEFAULT_TYPE_ATTRIBUTES, description=None):

        type_obj = type(self)
        if type_obj in ConfigType.__cache:
            check.failed(
                (
                    '{type_obj} already in cache. You **must** use the inst() class method '
                    'to construct ConfigTypes and not the ctor'.format(type_obj=type_obj)
                )
            )

        self.name = check.opt_str_param(name, 'name', type(self).__name__)
        self.description = check.opt_str_param(description, 'description')
        self.type_attributes = check.inst_param(
            type_attributes, 'type_attributes', ConfigTypeAttributes
        )

    __cache = {}

    @classmethod
    def inst(cls):
        if cls not in ConfigType.__cache:
            ConfigType.__cache[cls] = cls()
        return ConfigType.__cache[cls]

    @staticmethod
    def from_builtin_enum(builtin_enum):
        check.inst_param(builtin_enum, 'builtin_enum', BuiltinEnum)
        return _CONFIG_MAP[builtin_enum]

    @property
    def is_system_config(self):
        return self.type_attributes.is_system_config

    @property
    def is_builtin(self):
        return self.type_attributes.is_builtin

    @property
    def is_named(self):
        return self.type_attributes.is_named

    @property
    def has_fields(self):
        return self.is_composite or self.is_selector

    @property
    def is_scalar(self):
        return False

    @property
    def is_list(self):
        return False

    @property
    def is_nullable(self):
        return False

    @property
    def is_composite(self):
        return False

    @property
    def is_selector(self):
        return False

    @property
    def is_any(self):
        return False

    @property
    def inner_types(self):
        return []


# Scalars, Composites, Selectors, Lists, Nullable, Any


class ConfigScalar(ConfigType):
    @property
    def is_scalar(self):
        return True

    def is_config_scalar_valid(self, _config_value):
        check.not_implemented('must implement')


class ConfigList(ConfigType):
    def __init__(self, inner_type, *args, **kwargs):
        self.inner_type = check.inst_param(inner_type, 'inner_type', ConfigType)
        super(ConfigList, self).__init__(*args, **kwargs)

    def is_list(self):
        return True

    @property
    def inner_types(self):
        return [self.inner_type] + self.inner_type.inner_types


class ConfigNullable(ConfigType):
    def __init__(self, inner_type, *args, **kwargs):
        self.inner_type = check.inst_param(inner_type, 'inner_type', ConfigType)
        super(ConfigNullable, self).__init__(*args, **kwargs)

    @property
    def is_nullable(self):
        return True

    @property
    def inner_types(self):
        return [self.inner_type] + self.inner_type.inner_types


class ConfigAny(ConfigType):
    @property
    def is_any(self):
        return True


class BuiltinConfigScalar(ConfigScalar):
    def __init__(self, name=None, description=None):
        super(BuiltinConfigScalar, self).__init__(
            name=name,
            description=description,
            type_attributes=ConfigTypeAttributes(is_builtin=True),
        )


class Int(BuiltinConfigScalar):
    def __init__(self):
        super(Int, self).__init__(description='')

    def is_config_scalar_valid(self, config_value):
        return not isinstance(config_value, bool) and isinstance(config_value, six.integer_types)


class _StringishBuiltin(BuiltinConfigScalar):
    def is_config_scalar_valid(self, config_value):
        return isinstance(config_value, six.string_types)


class String(_StringishBuiltin):
    def __init__(self):
        super(String, self).__init__(description='')


class Path(_StringishBuiltin):
    def __init__(self):
        super(Path, self).__init__(description='')


class Bool(BuiltinConfigScalar):
    def __init__(self):
        super(Bool, self).__init__(description='')

    def is_config_scalar_valid(self, config_value):
        return isinstance(config_value, bool)


class Float(BuiltinConfigScalar):
    def __init__(self):
        super(Float, self).__init__(description='')

    def is_config_scalar_valid(self, config_value):
        return isinstance(config_value, float)


class Any(ConfigAny):
    def __init__(self):
        super(Any, self).__init__(type_attributes=ConfigTypeAttributes(is_builtin=True))


def Nullable(inner_type):
    check.inst_param(inner_type, 'inner_type', ConfigType)

    class _Nullable(ConfigNullable):
        def __init__(self):
            super(_Nullable, self).__init__(
                name='Nullable.{inner_type}'.format(inner_type=inner_type.name),
                type_attributes=ConfigTypeAttributes(is_builtin=True, is_named=False),
                inner_type=inner_type,
            )

    return _Nullable


def List(inner_type):
    check.inst_param(inner_type, 'inner_type', ConfigType)

    class _List(ConfigList):
        def __init__(self):
            super(_List, self).__init__(
                name='List.{inner_type}'.format(inner_type=inner_type.name),
                description='List of {inner_type}'.format(inner_type=inner_type.name),
                type_attributes=ConfigTypeAttributes(is_builtin=True, is_named=False),
                inner_type=inner_type,
            )

    return _List


def resolve_to_config_list(list_type):
    check.inst_param(list_type, 'list_type', WrappingListType)
    return List(resolve_to_config_type(list_type.inner_type))


def resolve_to_config_nullable(nullable_type):
    check.inst_param(nullable_type, 'nullable_type', WrappingNullableType)
    return Nullable(resolve_to_config_type(nullable_type.inner_type))


def resolve_to_config_type(dagster_type):
    if dagster_type is None:
        return ConfigAny.inst()
    if isinstance(dagster_type, BuiltinEnum):
        return ConfigType.from_builtin_enum(dagster_type)
    if isinstance(dagster_type, WrappingListType):
        return resolve_to_config_list(dagster_type).inst()
    if isinstance(dagster_type, WrappingNullableType):
        return resolve_to_config_nullable(dagster_type).inst()
    if issubclass(dagster_type, ConfigType):
        return dagster_type.inst()

    check.failed('should not reach')


_CONFIG_MAP = {
    BuiltinEnum.ANY: Any.inst(),
    BuiltinEnum.STRING: String.inst(),
    BuiltinEnum.INT: Int.inst(),
    BuiltinEnum.BOOL: Bool.inst(),
    BuiltinEnum.PATH: Path.inst(),
}



from dagster import check
from .runtime import PythonObjectType, RuntimeType


def create_inner_class(
    bare_cls, name, description, input_schema=None, output_schema=None, marshalling_strategy=None
):
    class _ObjectType(PythonObjectType):
        def __init__(self):
            super(_ObjectType, self).__init__(
                name=name,
                description=description,
                python_type=bare_cls,
                input_schema=input_schema,
                output_schema=output_schema,
                marshalling_strategy=marshalling_strategy,
            )

    type_inst = _ObjectType.inst()

    make_klass_runtime_type_decorated_klass(bare_cls, type_inst)
    return bare_cls


def dagster_type(name=None, description=None):
    def _with_args(bare_cls):
        check.type_param(bare_cls, 'bare_cls')
        return create_inner_class(
            bare_cls=bare_cls, name=name if name else bare_cls.__name__, description=description
        )

    # check for no args, no parens case
    if callable(name):
        klass = name
        return create_inner_class(bare_cls=klass, name=klass.__name__, description=None)

    return _with_args


MAGIC_RUNTIME_TYPE_NAME = '__runtime_type'


def is_runtime_type_decorated_klass(klass):
    check.type_param(klass, 'klass')
    return hasattr(klass, MAGIC_RUNTIME_TYPE_NAME)


def get_runtime_type_on_decorated_klass(klass):
    check.type_param(klass, 'klass')
    return getattr(klass, MAGIC_RUNTIME_TYPE_NAME)


def make_klass_runtime_type_decorated_klass(klass, runtime_type):
    check.type_param(klass, 'klass')
    check.inst_param(runtime_type, 'runtime_type', RuntimeType)
    setattr(klass, MAGIC_RUNTIME_TYPE_NAME, runtime_type)


def make_dagster_type(
    existing_type,
    name=None,
    description=None,
    input_schema=None,
    output_schema=None,
    marshalling_strategy=None,
):
    check.type_param(existing_type, 'existing_type')
    check.opt_str_param(name, 'name')
    check.opt_str_param(description, 'description')
    return create_inner_class(
        existing_type,
        name=existing_type.__name__ if name is None else name,
        description=description,
        input_schema=input_schema,
        output_schema=output_schema,
        marshalling_strategy=marshalling_strategy,
    )



from dagster import RepositoryDefinition

from .hello_world import define_hello_world_pipeline
from .hello_dag import define_hello_dag_pipeline
from .actual_dag import define_diamond_dag_pipeline
from .config import define_configurable_hello_pipeline
from .execution_context import define_execution_context_pipeline_step_two


def define_repository():
    return RepositoryDefinition(
        name='tutorial_repository',
        pipeline_dict={
            'hello_world_pipeline': define_hello_world_pipeline,
            'hello_dag_pipeline': define_hello_dag_pipeline,
            'actual_dag_pipeline': define_diamond_dag_pipeline,
            'configurable_hello_pipeline': define_configurable_hello_pipeline,
            'execution_context_pipeline': define_execution_context_pipeline_step_two,
        },
    )



from dagster import (
    DependencyDefinition,
    InputDefinition,
    OutputDefinition,
    PipelineDefinition,
    Result,
    ReentrantInfo,
    lambda_solid,
    solid,
    types,
)

from dagster.core.execution import (
    create_subplan,
    create_execution_plan,
    execute_plan,
    create_typed_environment,
    ExecutionPlanInfo,
    ExecutionPlanSubsetInfo,
    yield_context,
)

from dagster.core.execution_plan.utility import VALUE_OUTPUT


def define_two_int_pipeline():
    @lambda_solid
    def return_one():
        return 1

    @lambda_solid(inputs=[InputDefinition('num')])
    def add_one(num):
        return num + 1

    return PipelineDefinition(
        name='pipeline_ints',
        solids=[return_one, add_one],
        dependencies={'add_one': {'num': DependencyDefinition('return_one')}},
    )


def test_execution_plan_simple_two_steps():
    pipeline_def = define_two_int_pipeline()
    execution_plan = create_execution_plan(pipeline_def)

    assert isinstance(execution_plan.steps, list)
    assert len(execution_plan.steps) == 2

    assert execution_plan.get_step_by_key('return_one.transform')
    assert execution_plan.get_step_by_key('add_one.transform')

    step_results = execute_plan(pipeline_def, execution_plan)
    assert len(step_results) == 2

    assert step_results[0].step.key == 'return_one.transform'
    assert step_results[0].success
    assert step_results[0].success_data.value == 1

    assert step_results[1].step.key == 'add_one.transform'
    assert step_results[1].success
    assert step_results[1].success_data.value == 2


def test_create_subplan_source_step():
    pipeline_def = define_two_int_pipeline()
    typed_environment = create_typed_environment(pipeline_def, None)
    execution_plan = create_execution_plan(pipeline_def)
    with yield_context(pipeline_def, typed_environment) as context:
        subplan = create_subplan(
            ExecutionPlanInfo(
                context=context, pipeline=pipeline_def, environment=typed_environment
            ),
            execution_plan,
            ExecutionPlanSubsetInfo(['return_one.transform']),
        )
        assert subplan
        assert len(subplan.steps) == 1
        assert subplan.steps[0].key == 'return_one.transform'
        assert not subplan.steps[0].step_inputs
        assert len(subplan.steps[0].step_outputs) == 1
        assert len(subplan.topological_steps()) == 1


def test_create_subplan_middle_step():
    pipeline_def = define_two_int_pipeline()
    typed_environment = create_typed_environment(pipeline_def, None)
    execution_plan = create_execution_plan(pipeline_def)
    with yield_context(pipeline_def, typed_environment) as context:
        subplan = create_subplan(
            ExecutionPlanInfo(
                context=context, pipeline=pipeline_def, environment=typed_environment
            ),
            execution_plan,
            ExecutionPlanSubsetInfo(['add_one.transform'], {'add_one.transform': {'num': 2}}),
        )
        assert subplan
        steps = subplan.topological_steps()
        assert len(steps) == 2
        assert steps[0].key == 'add_one.transform.input.num.value'
        assert not steps[0].step_inputs
        assert len(steps[0].step_outputs) == 1
        assert steps[1].key == 'add_one.transform'
        assert len(steps[1].step_inputs) == 1
        step_input = steps[1].step_inputs[0]
        assert step_input.prev_output_handle.step.key == 'add_one.transform.input.num.value'
        assert step_input.prev_output_handle.output_name == VALUE_OUTPUT
        assert len(steps[1].step_outputs) == 1
        assert len(subplan.topological_steps()) == 2
        assert [step.key for step in subplan.topological_steps()] == [
            'add_one.transform.input.num.value',
            'add_one.transform',
        ]


def test_execution_plan_source_step():
    pipeline_def = define_two_int_pipeline()
    execution_plan = create_execution_plan(pipeline_def)
    step_results = execute_plan(
        pipeline_def,
        execution_plan,
        subset_info=ExecutionPlanSubsetInfo(included_steps=['return_one.transform']),
    )

    assert len(step_results) == 1
    assert step_results[0].success_data.value == 1


def test_execution_plan_middle_step():
    pipeline_def = define_two_int_pipeline()
    execution_plan = create_execution_plan(pipeline_def)
    step_results = execute_plan(
        pipeline_def,
        execution_plan,
        subset_info=ExecutionPlanSubsetInfo(
            ['add_one.transform'], {'add_one.transform': {'num': 2}}
        ),
    )

    assert len(step_results) == 2
    assert step_results[1].success_data.value == 3


def test_execution_plan_two_outputs():
    @solid(outputs=[OutputDefinition(types.Int, 'num_one'), OutputDefinition(types.Int, 'num_two')])
    def return_one_two(_info):
        yield Result(1, 'num_one')
        yield Result(2, 'num_two')

    pipeline_def = PipelineDefinition(name='return_one_two_pipeline', solids=[return_one_two])

    execution_plan = create_execution_plan(pipeline_def)

    step_results = execute_plan(pipeline_def, execution_plan)

    # FIXME: we should change this to be *single* result with two outputs
    assert step_results[0].step.key == 'return_one_two.transform'
    assert step_results[0].success_data.value == 1
    assert step_results[0].success_data.output_name == 'num_one'
    assert step_results[1].step.key == 'return_one_two.transform'
    assert step_results[1].success_data.value == 2
    assert step_results[1].success_data.output_name == 'num_two'


def test_reentrant_execute_plan():
    called = {}

    @solid
    def has_context_value(info):
        assert info.context.has_context_value('foo')
        assert info.context.get_context_value('foo') == 'bar'
        called['yup'] = True

    pipeline_def = PipelineDefinition(name='has_context_value_pipeline', solids=[has_context_value])
    execution_plan = create_execution_plan(pipeline_def)

    step_results = execute_plan(
        pipeline_def, execution_plan, reentrant_info=ReentrantInfo(context_stack={'foo': 'bar'})
    )

    assert called['yup']
    assert len(step_results) == 1



import pytest

from dagster import (
    DependencyDefinition,
    Dict,
    Field,
    InputDefinition,
    OutputDefinition,
    PipelineContextDefinition,
    PipelineDefinition,
    SolidInstance,
    String,
    lambda_solid,
    solid,
    types,
)
from dagster.core.errors import DagsterInvalidDefinitionError


def test_deps_equal():
    assert DependencyDefinition('foo') == DependencyDefinition('foo')
    assert DependencyDefinition('foo') != DependencyDefinition('bar')

    assert DependencyDefinition('foo', 'bar') == DependencyDefinition('foo', 'bar')
    assert DependencyDefinition('foo', 'bar') != DependencyDefinition('foo', 'quuz')


def test_pipeline_types():
    @lambda_solid
    def produce_string():
        return 'foo'

    @solid(
        inputs=[InputDefinition('input_one', types.String)],
        outputs=[OutputDefinition(types.Any)],
        config_field=Field(Dict({'another_field': Field(types.Int)})),
    )
    def solid_one(_info, input_one):
        raise Exception('should not execute')

    pipeline_def = PipelineDefinition(
        solids=[produce_string, solid_one],
        dependencies={'solid_one': {'input_one': DependencyDefinition('produce_string')}},
        context_definitions={
            'context_one': PipelineContextDefinition(
                context_fn=lambda: None, config_field=Field(Dict({'field_one': Field(String)}))
            )
        },
    )

    assert pipeline_def.has_config_type('String')
    assert pipeline_def.has_config_type('Int')
    assert not pipeline_def.has_config_type('SomeName')


def test_mapper_errors():
    @lambda_solid
    def solid_a():
        print('a: 1')
        return 1

    with pytest.raises(DagsterInvalidDefinitionError) as excinfo_1:
        PipelineDefinition(
            solids=[solid_a], dependencies={'solid_b': {'arg_a': DependencyDefinition('solid_a')}}
        )
    assert str(excinfo_1.value) == 'Solid solid_b in dependency dictionary not found in solid list'

    with pytest.raises(DagsterInvalidDefinitionError) as excinfo_2:
        PipelineDefinition(
            solids=[solid_a],
            dependencies={
                SolidInstance('solid_b', alias='solid_c'): {
                    'arg_a': DependencyDefinition('solid_a')
                }
            },
        )
    assert (
        str(excinfo_2.value)
        == 'Solid solid_b (aliased by solid_c in dependency dictionary) not found in solid list'
    )



from dagster import PipelineDefinition, OutputDefinition, lambda_solid, types


def test_materialization_schema_types():
    @lambda_solid(output=OutputDefinition(types.String))
    def return_one():
        return 1

    pipeline_def = PipelineDefinition(name='test_materialization_schema_types', solids=[return_one])

    string_mat_schema = pipeline_def.config_type_named('String.MaterializationSchema')

    string_json_mat_schema = string_mat_schema.fields['json'].config_type

    assert pipeline_def.config_type_named(string_json_mat_schema.name)



from dagster import check
from dagster.core.definitions import Result, Solid

from .objects import (
    ExecutionValueSubPlan,
    ExecutionStep,
    StepInput,
    StepOutput,
    StepOutputHandle,
    StepTag,
)

JOIN_OUTPUT = 'join_output'


def __join_lambda(_context, _step, inputs):
    yield Result(output_name=JOIN_OUTPUT, value=list(inputs.values())[0])


def create_join_step(solid, step_key, prev_steps, prev_output_name):
    check.inst_param(solid, 'solid', Solid)
    check.str_param(step_key, 'step_key')
    check.list_param(prev_steps, 'prev_steps', of_type=ExecutionStep)
    check.invariant(len(prev_steps) > 0)
    check.str_param(prev_output_name, 'output_name')

    step_inputs = []
    seen_runtime_type = None
    for prev_step in prev_steps:
        prev_step_output = prev_step.step_output_named(prev_output_name)

        if seen_runtime_type is None:
            seen_runtime_type = prev_step_output.runtime_type
        else:
            check.invariant(seen_runtime_type == prev_step_output.runtime_type)

        output_handle = StepOutputHandle(prev_step, prev_output_name)

        step_inputs.append(StepInput(prev_step.key, prev_step_output.runtime_type, output_handle))

    return ExecutionStep(
        key=step_key,
        step_inputs=step_inputs,
        step_outputs=[StepOutput(JOIN_OUTPUT, seen_runtime_type)],
        compute_fn=__join_lambda,
        tag=StepTag.JOIN,
        solid=solid,
    )


def create_joining_subplan(solid, join_step_key, parallel_steps, parallel_step_output):
    '''
    This captures a common pattern of fanning out a single value to N steps,
    where each step has similar structure. The strict requirement here is that each step
    must provide an output named the parameters parallel_step_output.

    This takes those steps and then uses a join node to coalesce them so that downstream
    steps can depend on a single output.

    Currently the join step just does a passthrough with no computation. It remains
    to be seen if there should be any work or verification done in this step, especially
    in multi-process environments that require marshalling between steps.
    '''
    check.inst_param(solid, 'solid', Solid)
    check.str_param(join_step_key, 'join_step_key')
    check.list_param(parallel_steps, 'parallel_steps', of_type=ExecutionStep)
    check.str_param(parallel_step_output, 'parallel_step_output')

    for parallel_step in parallel_steps:
        check.invariant(parallel_step.has_step_output(parallel_step_output))

    join_step = create_join_step(solid, join_step_key, parallel_steps, parallel_step_output)

    output_name = join_step.step_outputs[0].name
    return ExecutionValueSubPlan(
        parallel_steps + [join_step], StepOutputHandle(join_step, output_name)
    )


VALUE_OUTPUT = 'value_output'


def create_value_thunk_step(solid, runtime_type, step_key, value):
    def _fn(_context, _step, _inputs):
        yield Result(value, VALUE_OUTPUT)

    return StepOutputHandle(
        ExecutionStep(
            key=step_key,
            step_inputs=[],
            step_outputs=[StepOutput(VALUE_OUTPUT, runtime_type)],
            compute_fn=_fn,
            tag=StepTag.VALUE_THUNK,
            solid=solid,
        ),
        VALUE_OUTPUT,
    )



from collections import defaultdict

from dagster import (
    DependencyDefinition,
    InputDefinition,
    lambda_solid,
    OutputDefinition,
    PipelineDefinition,
    RepositoryDefinition,
    solid,
    types,
)


@solid(
    config_field=types.Field(types.Dict({'word': types.Field(types.String)}))
)
def double_the_word(info):
    return info.config['word'] * 2


@lambda_solid(inputs=[InputDefinition('word')])
def count_letters(word):
    counts = defaultdict(int)
    for letter in word:
        counts[letter] += 1
    return dict(counts)


@solid(
    config_field=types.Field(types.Dict({'word': types.Field(types.String)})),
    outputs=[OutputDefinition(types.String)],
)
def typed_double_the_word(info):
    return info.config['word'] * 2


@solid(
    config_field=types.Field(types.Dict({'word': types.Field(types.String)})),
    outputs=[OutputDefinition(types.Int)],
)
def typed_double_the_word_error(info):
    return info.config['word'] * 2


def define_demo_configuration_schema_pipeline():
    return PipelineDefinition(
        name='demo_configuration_schema',
        solids=[double_the_word, count_letters],
        dependencies={
            'count_letters': {'word': DependencyDefinition('double_the_word')}
        },
    )


def define_typed_demo_configuration_schema_pipeline():
    return PipelineDefinition(
        name='typed_demo_configuration_schema',
        solids=[typed_double_the_word, count_letters],
        dependencies={
            'count_letters': {
                'word': DependencyDefinition('typed_double_the_word')
            }
        },
    )


def define_typed_demo_configuration_schema_error_pipeline():
    return PipelineDefinition(
        name='typed_demo_configuration_schema_error',
        solids=[typed_double_the_word_error, count_letters],
        dependencies={
            'count_letters': {
                'word': DependencyDefinition('typed_double_the_word_error')
            }
        },
    )


def define_demo_configuration_schema_repo():
    return RepositoryDefinition(
        name='demo_configuration_schema_repo',
        pipeline_dict={
            'demo_configuration_schema': define_demo_configuration_schema_pipeline,
            'typed_demo_configuration_schema': define_typed_demo_configuration_schema_pipeline,
            'typed_demo_configuration_schema_error': define_typed_demo_configuration_schema_error_pipeline,
        },
    )



from dagster import check

from dagster.core.definitions import InputDefinition, Result, Solid

from dagster.core.errors import DagsterInvariantViolationError

from .objects import ExecutionPlanInfo, ExecutionStep, StepOutput, StepOutputHandle, StepTag

INPUT_THUNK_OUTPUT = 'input_thunk_output'


def _create_input_thunk_execution_step(solid, input_def, config_value):
    check.invariant(input_def.runtime_type.input_schema)

    def _fn(_context, _step, _inputs):
        value = input_def.runtime_type.input_schema.construct_from_config_value(config_value)
        yield Result(value, INPUT_THUNK_OUTPUT)

    return ExecutionStep(
        key=solid.name + '.' + input_def.name + '.input_thunk',
        step_inputs=[],
        step_outputs=[StepOutput(name=INPUT_THUNK_OUTPUT, runtime_type=input_def.runtime_type)],
        compute_fn=_fn,
        tag=StepTag.INPUT_THUNK,
        solid=solid,
    )


def create_input_thunk_execution_step(info, solid, input_def, value):
    check.inst_param(info, 'info', ExecutionPlanInfo)
    check.inst_param(solid, 'solid', Solid)
    check.inst_param(input_def, 'input_def', InputDefinition)

    dependency_structure = info.pipeline.dependency_structure
    input_handle = solid.input_handle(input_def.name)

    if dependency_structure.has_dep(input_handle):
        raise DagsterInvariantViolationError(
            (
                'In pipeline {pipeline_name} solid {solid_name}, input {input_name} '
                'you have specified an input via config while also specifying '
                'a dependency. Either remove the dependency, specify a subdag '
                'to execute, or remove the inputs specification in the environment.'
            ).format(
                pipeline_name=info.pipeline.name, solid_name=solid.name, input_name=input_def.name
            )
        )

    input_thunk = _create_input_thunk_execution_step(solid, input_def, value)
    return StepOutputHandle(input_thunk, INPUT_THUNK_OUTPUT)



from dagster import execute_pipeline

from ..execution_context import (
    define_execution_context_pipeline_step_one,
    define_execution_context_pipeline_step_two,
    define_execution_context_pipeline_step_three,
)


def test_execution_context():
    for pipeline_definition in [
        define_execution_context_pipeline_step_one,
        define_execution_context_pipeline_step_two,
        define_execution_context_pipeline_step_three,
    ]:
        pipeline = pipeline_definition()
        execute_pipeline(
            pipeline,
            {'context': {'default': {'config': {'log_level': 'DEBUG'}}}},
        )



from collections import namedtuple

from dagster import check
from dagster.core.execution_context import RuntimeExecutionContext

from .dependency import Solid
from .expectation import ExpectationDefinition
from .input import InputDefinition
from .output import OutputDefinition
from .pipeline import PipelineDefinition


class ContextCreationExecutionInfo(
    namedtuple('_ContextCreationExecutionInfo', 'config pipeline_def run_id')
):
    def __new__(cls, config, pipeline_def, run_id):
        return super(ContextCreationExecutionInfo, cls).__new__(
            cls,
            config,
            check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition),
            check.str_param(run_id, 'run_id'),
        )


class ExpectationExecutionInfo(
    namedtuple('_ExpectationExecutionInfo', 'context inout_def solid expectation_def')
):
    def __new__(cls, context, inout_def, solid, expectation_def):
        return super(ExpectationExecutionInfo, cls).__new__(
            cls,
            check.inst_param(context, 'context', RuntimeExecutionContext),
            check.inst_param(inout_def, 'inout_def', (InputDefinition, OutputDefinition)),
            check.inst_param(solid, 'solid', Solid),
            check.inst_param(expectation_def, 'expectation_def', ExpectationDefinition),
        )


class TransformExecutionInfo(
    namedtuple('_TransformExecutionInfo', 'context config solid pipeline_def')
):
    '''An instance of TransformExecutionInfo is passed every solid transform function.

    Attributes:

        context (ExecutionContext): Context instance for this pipeline invocation
        config (Any): Config object for current solid
    '''

    def __new__(cls, context, config, solid, pipeline_def):
        return super(TransformExecutionInfo, cls).__new__(
            cls,
            check.inst_param(context, 'context', RuntimeExecutionContext),
            config,
            check.inst_param(solid, 'solid', Solid),
            check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition),
        )

    @property
    def solid_def(self):
        return self.solid.definition



import six

from dagster import check

from dagster.core.errors import DagsterRuntimeCoercionError

from .builtin_enum import BuiltinEnum
from .builtin_config_schemas import define_builtin_scalar_output_schema
from .config import Int as ConfigInt
from .config import String as ConfigString
from .config import Any as ConfigAny
from .config import ConfigType
from .config import List as ConfigList
from .config import Nullable as ConfigNullable
from .config_schema import InputSchema, OutputSchema, make_input_schema
from .marshal import MarshallingStrategy
from .dagster_type import check_dagster_type_param
from .wrapping import WrappingListType, WrappingNullableType


def check_opt_config_cls_param(config_cls, param_name):
    if config_cls is None:
        return config_cls
    check.invariant(isinstance(config_cls, type))
    check.param_invariant(issubclass(config_cls, ConfigType), param_name)
    return config_cls


class RuntimeType(object):
    def __init__(
        self,
        name=None,
        description=None,
        input_schema=None,
        output_schema=None,
        marshalling_strategy=None,
    ):

        type_obj = type(self)
        if type_obj in RuntimeType.__cache:
            check.failed(
                (
                    '{type_obj} already in cache. You **must** use the inst() class method '
                    'to construct RuntimeType and not the ctor'.format(type_obj=type_obj)
                )
            )

        self.name = check.opt_str_param(name, 'name', type(self).__name__)
        self.description = check.opt_str_param(description, 'description')
        self.input_schema = check.opt_inst_param(input_schema, 'input_schema', InputSchema)
        self.output_schema = check.opt_inst_param(output_schema, 'output_schema', OutputSchema)
        self.marshalling_strategy = check.opt_inst_param(
            marshalling_strategy, 'marshalling_strategy', MarshallingStrategy
        )

    __cache = {}

    @classmethod
    def inst(cls):
        if cls not in RuntimeType.__cache:
            RuntimeType.__cache[cls] = cls()
        return RuntimeType.__cache[cls]

    @staticmethod
    def from_builtin_enum(builtin_enum):
        check.inst_param(builtin_enum, 'builtin_enum', BuiltinEnum)
        return _RUNTIME_MAP[builtin_enum]

    def coerce_runtime_value(self, value):
        return value

    def throw_if_false(self, fn, value):
        if not fn(value):
            raise DagsterRuntimeCoercionError(
                'Expected valid value for {type_name} but got {value}'.format(
                    type_name=self.name, value=repr(value)
                )
            )
        return value

    def throw_if_not_string(self, value):
        return self.throw_if_false(lambda v: isinstance(v, six.string_types), value)

    @property
    def is_any(self):
        return False

    @property
    def is_scalar(self):
        return False

    @property
    def is_list(self):
        return False

    @property
    def is_nullable(self):
        return False


class BuiltinScalarRuntimeType(RuntimeType):
    @property
    def is_scalar(self):
        return True


INT_INPUT_SCHEMA = make_input_schema(ConfigInt)
INT_OUTPUT_SCHEMA = define_builtin_scalar_output_schema('Int')


class Int(BuiltinScalarRuntimeType):
    def __init__(self):
        super(Int, self).__init__(input_schema=INT_INPUT_SCHEMA, output_schema=INT_OUTPUT_SCHEMA)

    def coerce_runtime_value(self, value):
        return self.throw_if_false(
            lambda v: not isinstance(v, bool) and isinstance(v, six.integer_types), value
        )


STRING_INPUT_SCHEMA = make_input_schema(ConfigString)
STRING_OUTPUT_SCHEMA = define_builtin_scalar_output_schema('String')


class String(BuiltinScalarRuntimeType):
    def __init__(self):
        super(String, self).__init__(
            input_schema=STRING_INPUT_SCHEMA, output_schema=STRING_OUTPUT_SCHEMA
        )

    def coerce_runtime_value(self, value):
        return self.throw_if_not_string(value)


class Path(String):
    pass


class Float(RuntimeType):
    # TODO
    # def __init__(self):
    #     super(String, self).__init__(
    #         input_schema_cls=config.String,
    #         output_schema_cls=define_builtin_scalar_output_schema(name='String'),
    #     )
    def coerce_runtime_value(self, value):
        return self.throw_if_false(lambda v: isinstance(v, float), value)


class Bool(RuntimeType):
    def coerce_runtime_value(self, value):
        return self.throw_if_false(lambda v: isinstance(v, bool), value)


class Any(RuntimeType):
    def __init__(self):
        super(Any, self).__init__(input_schema=make_input_schema(ConfigAny))

    @property
    def is_any(self):
        return True


class PythonObjectType(RuntimeType):
    def __init__(self, python_type, *args, **kwargs):
        super(PythonObjectType, self).__init__(*args, **kwargs)
        self.python_type = check.type_param(python_type, 'python_type')

    def coerce_runtime_value(self, value):
        return self.throw_if_false(lambda v: isinstance(v, self.python_type), value)


class NullableType(RuntimeType):
    def __init__(self, inner_type):
        super(NullableType, self).__init__(
            name='Nullable.' + inner_type.name,
            input_schema=make_input_schema(ConfigNullable(inner_type.input_schema.schema_type))
            if inner_type.input_schema
            else None,
        )
        self.inner_type = inner_type

    def coerce_runtime_value(self, value):
        return None if value is None else self.inner_type.coerce_runtime_value(value)

    @property
    def is_nullable(self):
        return True


class ListType(RuntimeType):
    def __init__(self, inner_type):
        super(ListType, self).__init__(
            name='List.' + inner_type.name,
            input_schema=make_input_schema(ConfigList(inner_type.input_schema.schema_type))
            if inner_type.input_schema
            else None,
        )
        self.inner_type = inner_type

    def coerce_runtime_value(self, value):
        value = self.throw_if_false(lambda v: isinstance(value, list), value)
        return [self.inner_type.coerce_runtime_value(item) for item in value]

    @property
    def is_list(self):
        return True


def Nullable(inner_type):
    check.inst_param(inner_type, 'inner_type', RuntimeType)

    class _Nullable(NullableType):
        def __init__(self):
            super(_Nullable, self).__init__(inner_type)

    return _Nullable.inst()


def List(inner_type):
    check.inst_param(inner_type, 'inner_type', RuntimeType)

    class _List(ListType):
        def __init__(self):
            super(_List, self).__init__(inner_type)

    return _List.inst()


class Stringish(RuntimeType):
    def is_scalar(self):
        return True

    def coerce_runtime_value(self, value):
        return self.throw_if_not_string(value)


_RUNTIME_MAP = {
    BuiltinEnum.ANY: Any.inst(),
    BuiltinEnum.STRING: String.inst(),
    BuiltinEnum.INT: Int.inst(),
    BuiltinEnum.BOOL: Bool.inst(),
    BuiltinEnum.PATH: Path.inst(),
}


def resolve_to_runtime_type(dagster_type):
    # circular dep
    from .decorator import is_runtime_type_decorated_klass, get_runtime_type_on_decorated_klass

    check_dagster_type_param(dagster_type, 'dagster_type', RuntimeType)

    if dagster_type is None:
        return Any.inst()
    if isinstance(dagster_type, BuiltinEnum):
        return RuntimeType.from_builtin_enum(dagster_type)
    if isinstance(dagster_type, WrappingListType):
        return resolve_to_runtime_list(dagster_type)
    if isinstance(dagster_type, WrappingNullableType):
        return resolve_to_runtime_nullable(dagster_type)
    if is_runtime_type_decorated_klass(dagster_type):
        return get_runtime_type_on_decorated_klass(dagster_type)
    if issubclass(dagster_type, RuntimeType):
        return dagster_type.inst()

    check.failed('should not reach')


def resolve_to_runtime_list(list_type):
    check.inst_param(list_type, 'list_type', WrappingListType)
    return List(resolve_to_runtime_type(list_type.inner_type))


def resolve_to_runtime_nullable(nullable_type):
    check.inst_param(nullable_type, 'nullable_type', WrappingNullableType)
    return Nullable(resolve_to_runtime_type(nullable_type.inner_type))



import os
import zipfile


def zip_folder(folder_path, output_path):
    contents = os.walk(folder_path)
    try:
        zip_file = zipfile.ZipFile(output_path, 'w', zipfile.ZIP_DEFLATED)
        for root, folders, files in contents:
            # Include all subfolders, including empty ones.
            for path in folders + files:
                absolute_path = os.path.join(root, path)
                zip_file.write(absolute_path)
    finally:
        zip_file.close()



import pytest

from dagster import (
    DagsterInvariantViolationError,
    DependencyDefinition,
    ExecutionContext,
    Field,
    InputDefinition,
    Int,
    OutputDefinition,
    PipelineContextDefinition,
    PipelineDefinition,
    lambda_solid,
    solid,
    types,
)

from dagster.utils.test import execute_solid


def test_single_solid_in_isolation():
    @lambda_solid
    def solid_one():
        return 1

    pipeline_def = PipelineDefinition(solids=[solid_one])

    result = execute_solid(pipeline_def, 'solid_one')
    assert result.success
    assert result.transformed_value() == 1


def test_single_solid_with_single():
    @lambda_solid
    def solid_one():
        return 1

    @lambda_solid(inputs=[InputDefinition(name='num')])
    def add_one_solid(num):
        return num + 1

    pipeline_def = PipelineDefinition(
        solids=[solid_one, add_one_solid],
        dependencies={'add_one_solid': {'num': DependencyDefinition('solid_one')}},
    )

    result = execute_solid(pipeline_def, 'add_one_solid', inputs={'num': 2})
    assert result.success
    assert result.transformed_value() == 3


def test_single_solid_with_multiple_inputs():
    @lambda_solid
    def solid_one():
        return 1

    @lambda_solid(inputs=[InputDefinition(name='num_one'), InputDefinition('num_two')])
    def add_solid(num_one, num_two):
        return num_one + num_two

    pipeline_def = PipelineDefinition(
        solids=[solid_one, add_solid],
        dependencies={
            'add_solid': {
                'num_one': DependencyDefinition('solid_one'),
                'num_two': DependencyDefinition('solid_one'),
            }
        },
    )

    result = execute_solid(pipeline_def, 'add_solid', inputs={'num_one': 2, 'num_two': 3})

    assert result.success
    assert result.transformed_value() == 5


def test_single_solid_with_config():
    ran = {}

    @solid(config_field=Field(Int))
    def check_config_for_two(info):
        assert info.config == 2
        ran['check_config_for_two'] = True

    pipeline_def = PipelineDefinition(solids=[check_config_for_two])
    result = execute_solid(
        pipeline_def,
        'check_config_for_two',
        environment={'solids': {'check_config_for_two': {'config': 2}}},
    )

    assert result.success
    assert ran['check_config_for_two']


def test_single_solid_with_context_config():
    ran = {'check_context_config_for_two': 0}

    @solid
    def check_context_config_for_two(info):
        assert info.context.resources == 2
        ran['check_context_config_for_two'] += 1

    pipeline_def = PipelineDefinition(
        solids=[check_context_config_for_two],
        context_definitions={
            'test_context': PipelineContextDefinition(
                config_field=Field(Int, is_optional=True, default_value=2),
                context_fn=lambda info: ExecutionContext(resources=info.config),
            )
        },
    )

    result = execute_solid(
        pipeline_def,
        'check_context_config_for_two',
        environment={'context': {'test_context': {'config': 2}}},
    )

    assert result.success
    assert ran['check_context_config_for_two'] == 1

    result = execute_solid(pipeline_def, 'check_context_config_for_two')

    assert result.success
    assert ran['check_context_config_for_two'] == 2


def test_single_solid_error():
    class SomeError(Exception):
        pass

    @lambda_solid
    def throw_error():
        raise SomeError()

    pipeline_def = PipelineDefinition(solids=[throw_error])

    with pytest.raises(SomeError):
        execute_solid(pipeline_def, 'throw_error')


def test_single_solid_type_checking_output_error():
    @lambda_solid(output=OutputDefinition(types.Int))
    def return_string():
        return 'ksjdfkjd'

    pipeline_def = PipelineDefinition(solids=[return_string])

    with pytest.raises(DagsterInvariantViolationError):
        execute_solid(pipeline_def, 'return_string')


def test_failing_solid_execute_solid():
    class ThisException(Exception):
        pass

    @lambda_solid
    def throw_an_error():
        raise ThisException('nope')

    pipeline_def = PipelineDefinition(solids=[throw_an_error])

    with pytest.raises(ThisException):
        execute_solid(pipeline_def, 'throw_an_error')



from dagster import execute_pipeline
from dagster.tutorials.utils import check_cli_execute_file_pipeline
from dagster.utils import script_relative_path

from ..hello_dag import define_hello_dag_pipeline


def test_intro_tutorial_hello_dag():
    pipeline = define_hello_dag_pipeline()

    result = execute_pipeline(pipeline)

    assert result.success
    assert len(result.result_list) == 2
    assert result.result_for_solid('solid_one').transformed_value() == 'foo'
    assert result.result_for_solid('solid_two').transformed_value() == 'foofoo'
    return result


def test_tutorial_cli_hello_dag():
    check_cli_execute_file_pipeline(
        script_relative_path('../hello_dag.py'), 'define_hello_dag_pipeline'
    )



import keyword
import re

from dagster import check

from dagster.core.errors import DagsterInvalidDefinitionError

from dagster.core.types.builtin_enum import BuiltinEnum
from dagster.core.types.runtime import RuntimeType, Any

DEFAULT_OUTPUT = 'result'

DISALLOWED_NAMES = set(
    [
        'context',
        'conf',
        'config',
        'meta',
        'arg_dict',
        'dict',
        'input_arg_dict',
        'output_arg_dict',
        'int',
        'str',
        'float',
        'bool',
        'input',
        'output',
        'type',
    ]
    + keyword.kwlist  # just disallow all python keywords
)


def check_valid_name(name):
    check.str_param(name, 'name')
    if name in DISALLOWED_NAMES:
        raise DagsterInvalidDefinitionError('{name} is not allowed'.format(name=name))

    regex = r'^[A-Za-z0-9_]+$'
    if not re.match(regex, name):
        raise DagsterInvalidDefinitionError(
            '{name} must be in regex {regex}'.format(name=name, regex=regex)
        )
    return name


def check_two_dim_dict(ddict, param_name, key_type=None, value_type=None):
    check.dict_param(ddict, param_name, key_type=key_type, value_type=dict)
    for sub_dict in ddict.values():
        check.dict_param(sub_dict, 'sub_dict', key_type=key_type, value_type=value_type)
    return ddict


def check_opt_two_dim_dict(ddict, param_name, key_type=None, value_type=None):
    ddict = check.opt_dict_param(ddict, param_name, key_type=key_type, value_type=dict)
    for sub_dict in ddict.values():
        check.dict_param(sub_dict, 'sub_dict', key_type=key_type, value_type=value_type)
    return ddict


def check_two_dim_str_dict(ddict, param_name, value_type):
    return check_two_dim_dict(ddict, param_name, key_type=str, value_type=value_type)


def check_opt_two_dim_str_dict(ddict, param_name, value_type):
    return check_opt_two_dim_dict(ddict, param_name, key_type=str, value_type=value_type)


def _kv_str(key, value):
    return '{key}="{value}"'.format(key=key, value=repr(value))


def struct_to_string(name, **kwargs):
    props_str = ', '.join([_kv_str(key, value) for key, value in kwargs.items()])
    return '{name}({props_str})'.format(name=name, props_str=props_str)


def check_runtime_cls_arg(runtime_cls, arg_name):
    if isinstance(runtime_cls, BuiltinEnum):
        return RuntimeType.from_builtin_enum(runtime_cls)
    if runtime_cls is None:
        return Any.inst()
    check.param_invariant(issubclass(runtime_cls, RuntimeType), arg_name)
    return runtime_cls.inst()



from __future__ import print_function
from dagster import check

from dagster.utils.indenting_printer import IndentingPrinter

from .config import ConfigType, resolve_to_config_type


def print_type(config_type, print_fn=print):
    check.inst_param(config_type, 'config_type', ConfigType)
    check.callable_param(print_fn, 'print_fn')

    printer = IndentingPrinter(printer=print_fn)
    _do_print(config_type, printer)
    printer.line('')


def _do_print(config_type, printer):
    if config_type.is_list:
        printer.append('[')
        _do_print(config_type.inner_type, printer)
        printer.append(']')
    elif config_type.is_nullable:
        _do_print(config_type.inner_type, printer)
        printer.append('?')
    elif config_type.has_fields:
        printer.line('{')
        with printer.with_indent():
            for name, field in sorted(config_type.fields.items()):
                if field.is_optional:
                    printer.append(name + '?: ')
                else:
                    printer.append(name + ': ')
                _do_print(field.config_type, printer)
                printer.line('')

        printer.append('}')
    elif config_type.is_named:
        printer.append(config_type.name)
    else:
        check.failed('not supported')


def print_type_to_string(dagster_type):
    config_type = resolve_to_config_type(dagster_type)
    prints = []

    def _push(text):
        prints.append(text)

    print_type(config_type, _push)

    return '\n'.join(prints)



import pytest

from dagster import PipelineConfigEvaluationError, execute_pipeline
from dagster.tutorials.utils import check_cli_execute_file_pipeline
from dagster.utils import script_relative_path

from ..inputs import (
    execute_with_another_world,
    define_hello_typed_inputs_pipeline,
)


def test_hello_inputs_parameterized_pipeline():
    result = execute_with_another_world()
    assert result.success
    solid_result = result.result_for_solid('add_hello_to_word')
    assert solid_result.transformed_value() == 'Hello, Mars!'


def test_hello_inputs_parameterized_cli_pipeline():
    check_cli_execute_file_pipeline(
        script_relative_path('../inputs.py'),
        'define_hello_inputs_pipeline',
        script_relative_path('../inputs_env.yml'),
    )


def test_hello_typed_inputs():
    with pytest.raises(PipelineConfigEvaluationError):
        execute_pipeline(
            define_hello_typed_inputs_pipeline(),
            {'solids': {'add_hello_to_word': {'inputs': {'word': 343}}}},
        )



# pylint: disable=W0622,W0614,W0401
import pytest

from dagster import (
    DagsterExpectationFailedError,
    DependencyDefinition,
    ExpectationDefinition,
    ExpectationResult,
    Field,
    InputDefinition,
    Int,
    OutputDefinition,
    PipelineDefinition,
    execute_pipeline,
    lambda_solid,
    solid,
)


@solid(
    config_field=Field(Int),
    outputs=[
        OutputDefinition(
            Int,
            expectations=[
                ExpectationDefinition(
                    name='check_positive',
                    expectation_fn=lambda _info, value: ExpectationResult(
                        success=value > 0
                    ),
                )
            ],
        )
    ],
)
def ingest_a(info):
    return info.config


@solid(config_field=Field(Int), outputs=[OutputDefinition(Int)])
def ingest_b(info):
    return info.config


@lambda_solid(
    inputs=[InputDefinition('num_one', Int), InputDefinition('num_two', Int)],
    output=OutputDefinition(Int),
)
def add_ints(num_one, num_two):
    return num_one + num_two


def define_part_ten_step_one_pipeline():
    return PipelineDefinition(
        name='part_ten_step_one_pipeline',
        solids=[ingest_a, ingest_b, add_ints],
        dependencies={
            'add_ints': {
                'num_one': DependencyDefinition('ingest_a'),
                'num_two': DependencyDefinition('ingest_b'),
            }
        },
    )


def test_intro_tutorial_part_ten_step_one():
    result = execute_pipeline(
        define_part_ten_step_one_pipeline(),
        {
            'context': {'default': {'config': {'log_level': 'DEBUG'}}},
            'solids': {'ingest_a': {'config': 2}, 'ingest_b': {'config': 3}},
        },
    )

    assert result.success


def define_failing_environment_config():
    return {
        'context': {'default': {'config': {'log_level': 'DEBUG'}}},
        'solids': {'ingest_a': {'config': -2}, 'ingest_b': {'config': 3}},
    }


def test_intro_tutorial_part_ten_step_two_fails_hard():
    with pytest.raises(DagsterExpectationFailedError):
        execute_pipeline(
            define_part_ten_step_one_pipeline(),
            define_failing_environment_config(),
        )


def test_intro_tutorial_part_ten_step_two_fails_soft():
    result = execute_pipeline(
        define_part_ten_step_one_pipeline(),
        define_failing_environment_config(),
        throw_on_error=False,
    )

    assert not result.success


if __name__ == '__main__':
    execute_pipeline(
        define_part_ten_step_one_pipeline(),
        {
            'context': {'default': {'config': {'log_level': 'DEBUG'}}},
            'solids': {'ingest_a': {'config': -2}, 'ingest_b': {'config': 3}},
            'expectations': {'evaluate': True},
        },
    )



from collections import namedtuple
from enum import Enum

from dagster import check

from dagster.core.errors import DagsterError

from .config import ConfigType
from .field import Field


class DagsterEvaluationErrorReason(Enum):
    RUNTIME_TYPE_MISMATCH = 'RUNTIME_TYPE_MISMATCH'
    MISSING_REQUIRED_FIELD = 'MISSING_REQUIRED_FIELD'
    FIELD_NOT_DEFINED = 'FIELD_NOT_DEFINED'
    SELECTOR_FIELD_ERROR = 'SELECTOR_FIELD_ERROR'


class FieldNotDefinedErrorData(namedtuple('_FieldNotDefinedErrorData', 'field_name')):
    def __new__(cls, field_name):
        return super(FieldNotDefinedErrorData, cls).__new__(
            cls, check.str_param(field_name, 'field_name')
        )


class MissingFieldErrorData(namedtuple('_MissingFieldErrorData', 'field_name field_def')):
    def __new__(cls, field_name, field_def):
        return super(MissingFieldErrorData, cls).__new__(
            cls,
            check.str_param(field_name, 'field_name'),
            check.inst_param(field_def, 'field_def', Field),
        )


class RuntimeMismatchErrorData(namedtuple('_RuntimeMismatchErrorData', 'config_type value_rep')):
    def __new__(cls, config_type, value_rep):
        return super(RuntimeMismatchErrorData, cls).__new__(
            cls,
            check.inst_param(config_type, 'config_type', ConfigType),
            check.str_param(value_rep, 'value_rep'),
        )


class SelectorTypeErrorData(namedtuple('_SelectorTypeErrorData', 'dagster_type incoming_fields')):
    def __new__(cls, dagster_type, incoming_fields):
        check.param_invariant(dagster_type.is_selector, 'dagster_type')
        return super(SelectorTypeErrorData, cls).__new__(
            cls, dagster_type, check.list_param(incoming_fields, 'incoming_fields', of_type=str)
        )


ERROR_DATA_TYPES = (
    FieldNotDefinedErrorData,
    MissingFieldErrorData,
    RuntimeMismatchErrorData,
    SelectorTypeErrorData,
)


class EvaluationStack(namedtuple('_EvaluationStack', 'config_type entries')):
    def __new__(cls, config_type, entries):
        return super(EvaluationStack, cls).__new__(
            cls,
            check.inst_param(config_type, 'config_type', ConfigType),
            check.list_param(entries, 'entries', of_type=EvaluationStackEntry),
        )

    @property
    def levels(self):
        return [
            entry.field_name
            for entry in self.entries
            if isinstance(entry, EvaluationStackPathEntry)
        ]

    @property
    def type_in_context(self):
        ttype = self.entries[-1].config_type if self.entries else self.config_type
        # TODO: This is the wrong place for this
        # Should have general facility for unwrapping named types
        if ttype.is_nullable:
            return ttype.inner_type
        else:
            return ttype


class EvaluationStackEntry:  # marker interface
    pass


class EvaluationStackPathEntry(
    namedtuple('_EvaluationStackEntry', 'field_name field_def'), EvaluationStackEntry
):
    def __new__(cls, field_name, field_def):
        return super(EvaluationStackPathEntry, cls).__new__(
            cls,
            check.str_param(field_name, 'field_name'),
            check.inst_param(field_def, 'field_def', Field),
        )

    @property
    def config_type(self):
        return self.field_def.config_type


class EvaluationStackListItemEntry(
    namedtuple('_EvaluationStackListItemEntry', 'config_type list_index'), EvaluationStackEntry
):
    def __new__(cls, config_type, list_index):
        check.int_param(list_index, 'list_index')
        check.param_invariant(list_index >= 0, 'list_index')
        return super(EvaluationStackListItemEntry, cls).__new__(
            cls, check.inst_param(config_type, 'config_type', ConfigType), list_index
        )


class EvaluationError(namedtuple('_EvaluationError', 'stack reason message error_data')):
    def __new__(cls, stack, reason, message, error_data):
        return super(EvaluationError, cls).__new__(
            cls,
            check.inst_param(stack, 'stack', EvaluationStack),
            check.inst_param(reason, 'reason', DagsterEvaluationErrorReason),
            check.str_param(message, 'message'),
            check.inst_param(error_data, 'error_data', ERROR_DATA_TYPES),
        )


def friendly_string_for_error(error):
    type_in_context = error.stack.type_in_context

    path_msg, path = _get_friendly_path_info(error)

    type_msg = _get_type_msg(error, type_in_context)

    if error.reason == DagsterEvaluationErrorReason.MISSING_REQUIRED_FIELD:
        return 'Missing required field "{field_name}"{type_msg} {path_msg}'.format(
            field_name=error.error_data.field_name, path_msg=path_msg, type_msg=type_msg
        )
    elif error.reason == DagsterEvaluationErrorReason.FIELD_NOT_DEFINED:
        return 'Undefined field "{field_name}"{type_msg} {path_msg}'.format(
            field_name=error.error_data.field_name, path_msg=path_msg, type_msg=type_msg
        )
    elif error.reason == DagsterEvaluationErrorReason.RUNTIME_TYPE_MISMATCH:
        return 'Type failure at path "{path}"{type_msg}. Got "{value_rep}".'.format(
            path=path, type_msg=type_msg, value_rep=error.error_data.value_rep
        )
    elif error.reason == DagsterEvaluationErrorReason.SELECTOR_FIELD_ERROR:
        if error.error_data.incoming_fields:
            return (
                'Specified more than one field at path "{path}". '
                'You can only specify one field at this level.'
            ).format(path=path)
        else:
            return (
                'You specified no fields at path "{path}". '
                'You must specify one and only one field at this level.'
            ).format(path=path)
    else:
        check.failed('not supported')


def _get_type_msg(error, type_in_context):
    if error.stack.type_in_context.is_system_config:
        return ''
    else:
        return ' on type "{type_name}"'.format(type_name=type_in_context.name)


def _get_friendly_path_info(error):
    if not error.stack.entries:
        path = ''
        path_msg = 'at document config root.'
    else:
        comps = ['root']
        for entry in error.stack.entries:
            if isinstance(entry, EvaluationStackPathEntry):
                comp = ':' + entry.field_name
                comps.append(comp)
            elif isinstance(entry, EvaluationStackListItemEntry):
                comps.append('[{i}]'.format(i=entry.list_index))
            else:
                check.failed('unsupported')

        path = ''.join(comps)
        path_msg = 'at path ' + path
    return path_msg, path


class DagsterEvaluateConfigValueError(DagsterError):
    '''Indicates invalid value was passed to a type's evaluate_value method'''

    def __init__(self, stack, *args, **kwargs):
        super(DagsterEvaluateConfigValueError, self).__init__(*args, **kwargs)
        self.stack = check.inst_param(stack, 'stack', EvaluationStack)


class EvaluateValueResult(namedtuple('_EvaluateValueResult', 'success value errors')):
    def __new__(cls, success, value, errors):
        return super(EvaluateValueResult, cls).__new__(
            cls,
            check.bool_param(success, 'success'),
            value,
            check.list_param(errors, 'errors', of_type=EvaluationError),
        )

    def errors_at_level(self, *levels):
        return list(self._iterate_errors_at_level(list(levels)))

    def _iterate_errors_at_level(self, levels):
        check.list_param(levels, 'levels', of_type=str)
        for error in self.errors:
            if error.stack.levels == levels:
                yield error


def stack_with_field(stack, field_name, field_def):
    return EvaluationStack(
        config_type=stack.config_type,
        entries=stack.entries + [EvaluationStackPathEntry(field_name, field_def)],
    )


def stack_with_list_index(stack, list_index):
    list_type = stack.type_in_context
    check.invariant(list_type.is_list)
    return EvaluationStack(
        config_type=stack.config_type,
        entries=stack.entries + [EvaluationStackListItemEntry(list_type.inner_type, list_index)],
    )


def hard_create_config_value(config_type, config_value):
    check.inst_param(config_type, 'config_type', ConfigType)
    result = evaluate_config_value(config_type, config_value)
    check.invariant(result.success)
    return result.value


def evaluate_config_value(config_type, config_value):
    check.inst_param(config_type, 'config_type', ConfigType)
    errors = validate_config(config_type, config_value)
    if errors:
        return EvaluateValueResult(success=False, value=None, errors=errors)

    value = deserialize_config(config_type, config_value)

    return EvaluateValueResult(success=True, value=value, errors=[])


def validate_config(config_type, config_value):
    check.inst_param(config_type, 'config_type', ConfigType)

    return list(
        _validate_config(
            config_type, config_value, EvaluationStack(config_type=config_type, entries=[])
        )
    )


def _validate_config(config_type, config_value, stack):
    check.inst_param(config_type, 'config_type', ConfigType)
    check.inst_param(stack, 'stack', EvaluationStack)

    if config_type.is_scalar:
        if not config_type.is_config_scalar_valid(config_value):
            yield EvaluationError(
                stack=stack,
                reason=DagsterEvaluationErrorReason.RUNTIME_TYPE_MISMATCH,
                message='Value {value} is not valid for type {type_name}'.format(
                    value=config_value, type_name=config_type.name
                ),
                error_data=RuntimeMismatchErrorData(
                    config_type=config_type, value_rep=repr(config_value)
                ),
            )
        return

    errors = []

    if config_type.is_any:
        # no-op: we're safe
        return
    elif config_type.is_selector:
        errors = validate_selector_config_value(config_type, config_value, stack)
    elif config_type.is_composite:
        errors = validate_composite_config_value(config_type, config_value, stack)
    elif config_type.is_list:
        errors = validate_list_value(config_type, config_value, stack)
    elif config_type.is_nullable:
        errors = (
            []
            if config_value is None
            else _validate_config(config_type.inner_type, config_value, stack)
        )
    else:
        check.failed('Unsupported type {name}'.format(name=config_type.name))

    for error in errors:
        yield error


def deserialize_config(config_type, config_value):
    check.inst_param(config_type, 'config_type', ConfigType)
    return _deserialize_config(config_type, config_value)


def _deserialize_config(config_type, config_value):
    if config_type.is_scalar:
        return config_value
    elif config_type.is_selector:
        return deserialize_selector_config(config_type, config_value)
    elif config_type.is_composite:
        return deserialize_composite_config_value(config_type, config_value)
    elif config_type.is_list:
        return deserialize_list_value(config_type, config_value)
    elif config_type.is_nullable:
        if config_value is None:
            return None
        return _deserialize_config(config_type.inner_type, config_value)
    elif config_type.is_any:
        return config_value
    else:
        check.failed('Unsupported type {name}'.format(name=config_type.name))


## Selectors


def single_item(ddict):
    check.dict_param(ddict, 'ddict')
    check.param_invariant(len(ddict) == 1, 'ddict')
    return list(ddict.items())[0]


def validate_selector_config_value(selector_type, config_value, stack):
    check.param_invariant(selector_type.is_selector, 'selector_type')
    check.inst_param(stack, 'stack', EvaluationStack)

    if config_value and not isinstance(config_value, dict):
        yield EvaluationError(
            stack=stack,
            reason=DagsterEvaluationErrorReason.RUNTIME_TYPE_MISMATCH,
            message='Value for selector type {type_name} must be a dict got {value}'.format(
                type_name=selector_type.name, value=config_value
            ),
            error_data=RuntimeMismatchErrorData(
                config_type=selector_type, value_rep=repr(config_value)
            ),
        )
        return

    if config_value and len(config_value) > 1:
        incoming_fields = sorted(list(config_value.keys()))
        defined_fields = sorted(list(selector_type.fields.keys()))
        yield EvaluationError(
            stack=stack,
            reason=DagsterEvaluationErrorReason.SELECTOR_FIELD_ERROR,
            message=(
                'You can only specify a single field. You specified {incoming_fields}. '
                'The available fields are {defined_fields}'
            ).format(incoming_fields=incoming_fields, defined_fields=defined_fields),
            error_data=SelectorTypeErrorData(
                dagster_type=selector_type, incoming_fields=incoming_fields
            ),
        )
        return

    elif not config_value:
        defined_fields = sorted(list(selector_type.fields.keys()))
        if len(selector_type.fields) > 1:
            yield EvaluationError(
                stack=stack,
                reason=DagsterEvaluationErrorReason.SELECTOR_FIELD_ERROR,
                message=(
                    'Must specify a field if more one defined. Defined fields: ' '{defined_fields}'
                ).format(defined_fields=defined_fields),
                error_data=SelectorTypeErrorData(dagster_type=selector_type, incoming_fields=[]),
            )
            return

        field_name, field_def = single_item(selector_type.fields)

        if not field_def.is_optional:
            yield EvaluationError(
                stack=stack,
                reason=DagsterEvaluationErrorReason.SELECTOR_FIELD_ERROR,
                message=(
                    'Must specify the required field. Defined fields: ' '{defined_fields}'
                ).format(defined_fields=defined_fields),
                error_data=SelectorTypeErrorData(dagster_type=selector_type, incoming_fields=[]),
            )
            return

        incoming_field_value = field_def.default_value if field_def.default_provided else None

    else:
        check.invariant(config_value and len(config_value) == 1)

        field_name, incoming_field_value = single_item(config_value)
        if field_name not in selector_type.fields:
            yield create_field_not_defined_error(
                selector_type, stack, set(selector_type.fields.keys()), field_name
            )
            return

    parent_field = selector_type.fields[field_name]
    for error in _validate_config(
        parent_field.config_type,
        incoming_field_value,
        stack_with_field(stack, field_name, parent_field),
    ):
        yield error


def deserialize_selector_config(selector_type, config_value):
    check.param_invariant(selector_type.is_selector, 'selector_type')

    if config_value:
        check.invariant(config_value and len(config_value) == 1)
        field_name, incoming_field_value = single_item(config_value)

    else:
        field_name, field_def = single_item(selector_type.fields)
        incoming_field_value = field_def.default_value if field_def.default_provided else None

    parent_field = selector_type.fields[field_name]
    field_value = _deserialize_config(parent_field.config_type, incoming_field_value)
    return {field_name: field_value}


## Composites


def validate_composite_config_value(composite_type, config_value, stack):
    check.inst_param(composite_type, 'composite_type', ConfigType)
    check.param_invariant(composite_type.is_composite, 'composite_type')
    check.inst_param(stack, 'stack', EvaluationStack)

    if config_value and not isinstance(config_value, dict):
        yield EvaluationError(
            stack=stack,
            reason=DagsterEvaluationErrorReason.RUNTIME_TYPE_MISMATCH,
            message='Value for composite type {type_name} must be a dict got {value}'.format(
                type_name=composite_type.name, value=config_value
            ),
            error_data=RuntimeMismatchErrorData(
                config_type=composite_type, value_rep=repr(config_value)
            ),
        )
        return

    # ASK: this can crash on user error
    config_value = check.opt_dict_param(config_value, 'incoming_value', key_type=str)

    fields = composite_type.fields

    defined_fields = set(fields.keys())
    incoming_fields = set(config_value.keys())

    for received_field in incoming_fields:
        if received_field not in defined_fields:
            yield create_field_not_defined_error(
                composite_type, stack, defined_fields, received_field
            )

    for expected_field, field_def in fields.items():
        if expected_field in incoming_fields:
            for error in _validate_config(
                field_def.config_type,
                config_value[expected_field],
                stack_with_field(stack, expected_field, field_def),
            ):
                yield error

        elif field_def.is_optional:
            pass

        else:
            check.invariant(not field_def.default_provided)
            yield create_missing_required_field_error(
                composite_type, stack, defined_fields, expected_field
            )


def deserialize_composite_config_value(composite_type, config_value):
    check.param_invariant(composite_type.is_composite, 'composite_type')

    # ASK: this can crash on user error
    config_value = check.opt_dict_param(config_value, 'config_value', key_type=str)

    fields = composite_type.fields
    incoming_fields = set(config_value.keys())

    processed_fields = {}

    for expected_field, field_def in fields.items():
        if expected_field in incoming_fields:
            processed_fields[expected_field] = _deserialize_config(
                field_def.config_type, config_value[expected_field]
            )

        elif field_def.default_provided:
            processed_fields[expected_field] = field_def.default_value

        elif not field_def.is_optional:
            check.failed('Missing non-optional composite member not caught in validation')

    return processed_fields


## Lists


def validate_list_value(list_type, config_value, stack):
    check.param_invariant(list_type.is_list, 'list_type')
    check.inst_param(stack, 'stack', EvaluationStack)

    if not isinstance(config_value, list):
        yield EvaluationError(
            stack=stack,
            reason=DagsterEvaluationErrorReason.RUNTIME_TYPE_MISMATCH,
            message='Value for list type {type_name} must be a list got {value}'.format(
                type_name=list_type.name, value=config_value
            ),
            error_data=RuntimeMismatchErrorData(
                config_type=list_type, value_rep=repr(config_value)
            ),
        )
        return

    for index, item in enumerate(config_value):
        for error in _validate_config(
            list_type.inner_type, item, stack_with_list_index(stack, index)
        ):
            yield error


def deserialize_list_value(list_type, config_value):
    check.param_invariant(list_type.is_list, 'list_type')

    if not config_value:
        return []

    return [_deserialize_config(list_type.inner_type, item) for item in config_value]


##


def create_field_not_defined_error(config_type, stack, defined_fields, received_field):
    check.param_invariant(config_type.has_fields, 'config_type')
    return EvaluationError(
        stack=stack,
        reason=DagsterEvaluationErrorReason.FIELD_NOT_DEFINED,
        message='Field "{received}" is not defined on "{type_name}" Defined {defined}'.format(
            type_name=config_type.name, defined=repr(defined_fields), received=received_field
        ),
        error_data=FieldNotDefinedErrorData(field_name=received_field),
    )


def create_missing_required_field_error(config_type, stack, defined_fields, expected_field):
    check.param_invariant(config_type.has_fields, 'config_type')
    return EvaluationError(
        stack=stack,
        reason=DagsterEvaluationErrorReason.MISSING_REQUIRED_FIELD,
        message=(
            'Missing required field "{expected}" on "{type_name}". ' 'Defined fields: {defined}'
        ).format(expected=expected_field, type_name=config_type.name, defined=repr(defined_fields)),
        error_data=MissingFieldErrorData(
            field_name=expected_field, field_def=config_type.fields[expected_field]
        ),
    )



from collections import defaultdict

from dagster import (
    DependencyDefinition,
    Dict,
    Field,
    InputDefinition,
    PipelineDefinition,
    RepositoryDefinition,
    String,
    lambda_solid,
    solid,
)


@solid(config_field=Field(Dict({'word': Field(String)})))
def double_the_word(info):
    return info.config['word'] * 2


@lambda_solid(inputs=[InputDefinition('word')])
def count_letters(word):
    counts = defaultdict(int)
    for letter in word:
        counts[letter] += 1
    return dict(counts)


def define_demo_execution_pipeline():
    return PipelineDefinition(
        name='demo_execution',
        solids=[double_the_word, count_letters],
        dependencies={
            'count_letters': {'word': DependencyDefinition('double_the_word')}
        },
    )


def define_demo_execution_repo():
    return RepositoryDefinition(
        name='demo_execution_repo',
        pipeline_dict={'part_seven': define_demo_execution_pipeline},
    )
